# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides the ruler widget for the PyAnsys Geometry plotter."""

# First, verify graphics are available
try:
    from ansys.geometry.core.misc.checks import run_if_graphics_required

    run_if_graphics_required()
except ImportError as err:  # pragma: no cover
    raise err

from pathlib import Path

from ansys.tools.visualization_interface.backends.pyvista import PyVistaBackend
from ansys.tools.visualization_interface.backends.pyvista.widgets import PlotterWidget
from vtk import vtkButtonWidget, vtkPNGReader

from ansys.geometry.core.designer.designpoint import DesignPoint


class ShowDesignPoints(PlotterWidget):
    """Provides the a button to hide/show DesignPoint objects in the plotter.

    Parameters
    ----------
    plotter_helper : GeometryPlotter
        Provides the plotter to add the button to.
    """

    def __init__(self, plotter_helper: "PyVistaBackend") -> None:
        """Initialize the ``ShowDesignPoints`` class."""
        # Call PlotterWidget ctor
        super().__init__(plotter_helper._backend.pv_interface.scene)
        self.plotter_helper = plotter_helper

        # Initialize variables
        self._object_to_actors_map = self.plotter_helper._backend.pv_interface.object_to_actors_map
        self._button: vtkButtonWidget = (
            self.plotter_helper._backend._pl.scene.add_checkbox_button_widget(
                self.callback, position=(5, 438), size=30, border_size=3
            )
        )

    def callback(self, state: bool) -> None:
        """Remove or add the DesignPoint actors upon click.

        Parameters
        ----------
        state : bool
            State of the button, which is inherited from PyVista. The value is ``True``
            if the button is active.
        """
        if not state:
            for actor, object in self._object_to_actors_map.items():
                if isinstance(object, DesignPoint):
                    self.plotter_helper._backend._pl.scene.add_actor(actor)
        else:
            for actor, object in self._object_to_actors_map.items():
                if isinstance(object, DesignPoint):
                    self.plotter_helper._backend._pl.scene.remove_actor(actor)

    def update(self) -> None:
        """Define the configuration and representation of the button widget."""
        vr = self._button.GetRepresentation()
        icon_file = Path(__file__).parent / "_images" / "designpoint.png"
        representation = vtkPNGReader()
        representation.SetFileName(icon_file)
        representation.Update()
        image = representation.GetOutput()
        vr.SetButtonTexture(0, image)
        vr.SetButtonTexture(1, image)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides dataclasses for mating conditions."""

from dataclasses import dataclass


@dataclass
class MatingCondition:
    """MatingCondition dataclass."""

    id: str  # Id of the mating condition corresponding to the server-side object
    is_deleted: bool  # True if the condition has been deleted
    is_enabled: bool  # True if the condition is currently enabled
    is_satisfied: bool  # True if the condition has been met


@dataclass
class AlignCondition(MatingCondition):
    """AlignCondition dataclass."""

    offset: float  # Separation value (only defined if at least one geometric has a surface)
    is_reversed: bool  # True if the condition is reversed in sense
    is_valid: bool  # True if the condition is valid


@dataclass
class TangentCondition(MatingCondition):
    """TangentCondition dataclass."""

    offset: float  # Separation value (only if both are surfaces)
    is_reversed: bool  # True if the condition is reversed in sense
    is_valid: bool  # True if the condition is valid


@dataclass
class OrientCondition(MatingCondition):
    """OrientCondition dataclass."""

    offset: float  # Rotation angle
    is_reversed: bool  # True if the condition is reversed in sense
    is_valid: bool  # True if the condition is valid



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Trimmed curve class."""

from pint import Quantity

from ansys.geometry.core.connection.client import GrpcClient
from ansys.geometry.core.math.matrix import Matrix44
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D, Vector3D
from ansys.geometry.core.misc.measurements import Angle, Distance
from ansys.geometry.core.shapes.curves.curve import Curve
from ansys.geometry.core.shapes.curves.curve_evaluation import CurveEvaluation
from ansys.geometry.core.shapes.parameterization import Interval
from ansys.geometry.core.typing import Real


class TrimmedCurve:
    """Represents a trimmed curve.

    A trimmed curve is a curve that has a boundary. This boundary comes in the form of an interval.

    Parameters
    ----------
    geometry : Curve
        Underlying mathematical representation of the curve.
    start : Point3D
        Start point of the curve.
    end : Point3D
        End point of the curve.
    interval : Interval
        Parametric interval that bounds the curve.
    length : Quantity
        Length of the curve.
    grpc_client : GrpcClient, default: None
        gRPC client that is required for advanced functions such as ``intersect_curve()``.
    """

    def __init__(
        self,
        geometry: Curve,
        start: Point3D,
        end: Point3D,
        interval: Interval,
        length: Quantity,
        grpc_client: GrpcClient = None,
    ):
        """Initialize ``TrimmedCurve`` class."""
        self._geometry = geometry
        self._start = start
        self._end = end
        self._interval = interval
        self._length = length
        self._grpc_client = grpc_client

    @property
    def geometry(self) -> Curve:
        """Underlying mathematical curve."""
        return self._geometry

    @property
    def start(self) -> Point3D:
        """Start point of the curve."""
        return self._start

    @property
    def end(self) -> Point3D:
        """End point of the curve."""
        return self._end

    @property
    def length(self) -> Quantity:
        """Calculated length of the edge."""
        return self._length

    @property
    def interval(self) -> Interval:
        """Interval of the curve that provides its boundary."""
        return self._interval

    def evaluate_proportion(self, param: Real) -> CurveEvaluation:
        """Evaluate the curve at a proportional value.

        A parameter of ``0`` corresponds to the start of the curve, while a parameter of ``1``
        corresponds to the end of the curve.

        Parameters
        ----------
        param : Real
            Parameter in the proportional range [0,1].

        Returns
        -------
        CurveEvaluation
            Resulting curve evaluation.
        """
        bounds = self.interval
        return self.geometry.evaluate(bounds.start + bounds.get_span() * param)

    def intersect_curve(self, other: "TrimmedCurve") -> list[Point3D]:
        """Get the intersect points of this trimmed curve with another one.

        If the two trimmed curves do not intersect, an empty list is returned.

        Parameters
        ----------
        other : TrimmedCurve
            Trimmed curve to intersect with.

        Returns
        -------
        list[Point3D]
            All points of intersection between the curves.
        """
        if self._grpc_client is None:
            raise ValueError(
                """Because this trimmed curve was not initialized with a gRPC client,
                the method cannot be called."""
            )

        response = self._grpc_client.services.curves.intersect_curves(first=self, second=other)

        return [] if response.get("intersect") is False else response.get("points")

    def transformed_copy(self, matrix: Matrix44) -> "TrimmedCurve":
        """Return a copy of the trimmed curve transformed by the given matrix.

        Parameters
        ----------
        matrix : Matrix44
            Transformation matrix to apply to the curve.

        Returns
        -------
        TrimmedCurve
            A new trimmed curve with the transformation applied.
        """
        transformed_geometry = self.geometry.transformed_copy(matrix)
        transformed_start = self.start.transform(matrix)
        transformed_end = self.end.transform(matrix)

        return TrimmedCurve(
            transformed_geometry,
            transformed_start,
            transformed_end,
            self.interval,
            self.length,
        )

    def translate(self, direction: UnitVector3D, distance: Real | Quantity | Distance) -> None:
        """Translate the trimmed curve by a given vector and distance.

        Parameters
        ----------
        direction : UnitVector3D
            Direction of translation.
        distance : Real | Quantity | Distance
            Distance to translate.
        """
        distance = distance if isinstance(distance, Distance) else Distance(distance)
        translation_matrix = Matrix44.create_translation(direction * distance.value.m)

        translated_copy = self.transformed_copy(translation_matrix)

        # Update the current instance with the translated copy
        self._geometry = translated_copy.geometry
        self._start = translated_copy.start
        self._end = translated_copy.end
        self._length = translated_copy.length
        self._interval = translated_copy.interval

    def rotate(self, origin: Point3D, axis: UnitVector3D, angle: Real | Quantity | Angle) -> None:
        """Rotate the trimmed curve around a given axis centered at a given point.

        Parameters
        ----------
        origin : Point3D
            Origin point of the rotation.
        axis : UnitVector3D
            Axis of rotation.
        angle : Real | Quantity | Angle
            Angle to rotate in radians.
        """
        angle = angle if isinstance(angle, Angle) else Angle(angle)

        # Translate the curve to the origin
        translate_to_origin_matrix = Matrix44.create_translation(
            Vector3D([-origin.x.m, -origin.y.m, -origin.z.m])
        )
        translated_copy = self.transformed_copy(translate_to_origin_matrix)

        # Rotate the curve around the axis
        rotation_matrix = Matrix44.create_matrix_from_rotation_about_axis(axis, angle.value.m)
        rotated_copy = translated_copy.transformed_copy(rotation_matrix)

        # Translate the curve back to its original position
        translate_back_matrix = Matrix44.create_translation(
            Vector3D([origin.x.m, origin.y.m, origin.z.m])
        )
        translated_back_copy = rotated_copy.transformed_copy(translate_back_matrix)

        # Update the current instance with the rotated copy
        self._geometry = translated_back_copy.geometry
        self._start = translated_back_copy.start
        self._end = translated_back_copy.end
        self._length = translated_back_copy.length
        self._interval = translated_back_copy.interval

    def __repr__(self) -> str:
        """Represent the trimmed curve as a string."""
        return (
            f"TrimmedCurve(geometry: {type(self.geometry)}, start: {self.start}, end: {self.end}, "
            f"interval: {self.interval}, length: {self.length})"
        )


class ReversedTrimmedCurve(TrimmedCurve):
    """Represents a reversed trimmed curve.

    When a curve is reversed, its start and end points are swapped, and parameters for evaluations
    are handled to provide expected results conforming to the direction of the curve. For example,
    evaluating a trimmed curve proportionally at 0 evaluates at the start point of the curve,
    but evaluating a reversed trimmed curve proportionally at 0 evaluates at what was previously
    the end point of the curve but is now the start point.

    Parameters
    ----------
    geometry : Curve
        Underlying mathematical representation of the curve.
    start : Point3D
        Original start point of the curve.
    end : Point3D
        Original end point of the curve.
    interval : Interval
        Parametric interval that bounds the curve.
    length : Quantity
        Length of the curve.
    grpc_client : GrpcClient, default: None
        gRPC client that is required for advanced functions such as `intersect_curve()`.
    """

    def __init__(
        self,
        geometry: Curve,
        start: Point3D,
        end: Point3D,
        interval: Interval,
        length: Quantity,
        grpc_client: GrpcClient = None,
    ):
        """Initialize an instance of a reversed trimmed curve."""
        super().__init__(geometry, start, end, interval, length, grpc_client)

        # Swap start and end points
        temp = self._start
        self._start = self._end
        self._end = temp

    def evaluate_proportion(self, param: Real) -> CurveEvaluation:  # noqa: D102
        # Evaluate starting from the end
        return self.geometry.evaluate(self.interval.end - self.interval.get_span() * param)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module providing definitions for the backend types."""

from enum import Enum, unique
from typing import Union


@unique
class BackendType(Enum):
    """Provides an enum holding the available backend types."""

    DISCOVERY = 0
    SPACECLAIM = 1
    WINDOWS_SERVICE = 2
    LINUX_SERVICE = 3
    CORE_WINDOWS = 4
    CORE_LINUX = 5
    DISCOVERY_HEADLESS = 6

    @staticmethod
    def is_core_service(backend_type: "BackendType") -> bool:
        """Determine whether the backend is CoreService based or not.

        Parameters
        ----------
        backend_type : BackendType
            The backend type to check whether or not it's a CoreService type.

        Returns
        -------
        bool
            True if the backend is CoreService based, False otherwise.
        """
        return backend_type in (
            BackendType.LINUX_SERVICE,
            BackendType.CORE_WINDOWS,
            BackendType.CORE_LINUX,
        )

    @staticmethod
    def is_headless_service(backend_type: "BackendType") -> bool:
        """Determine whether the backend is a headless service or not.

        Parameters
        ----------
        backend_type : BackendType
            The backend type to check whether or not it's a headless service.

        Returns
        -------
        bool
            True if the backend is a headless service, False otherwise.
        """
        return backend_type in (
            BackendType.WINDOWS_SERVICE,
            BackendType.LINUX_SERVICE,
            BackendType.CORE_WINDOWS,
            BackendType.CORE_LINUX,
        )

    @staticmethod
    def is_linux_service(backend_type: "BackendType") -> bool:
        """Determine whether the backend is a Linux service or not.

        Parameters
        ----------
        backend_type : BackendType
            The backend type to check whether or not it's running on Linux.

        Returns
        -------
        bool
            True if the backend is running on Linux, False otherwise.
        """
        return backend_type in (
            BackendType.LINUX_SERVICE,
            BackendType.CORE_LINUX,
        )


class ApiVersions(Enum):
    """Provides an enum for all the compatibles API versions."""

    V_21 = 21
    V_22 = 22
    V_231 = 231
    V_232 = 232
    V_241 = 241
    V_242 = 242
    V_251 = 251
    V_252 = 252
    V_261 = 261
    V_271 = LATEST = 271

    @staticmethod
    def parse_input(version: Union[int, str, "ApiVersions"]) -> "ApiVersions":
        """Convert an input to an ApiVersions enum.

        Parameters
        ----------
        version : int | str | ApiVersions
            The version to convert to an ApiVersions enum.

        Returns
        -------
        ApiVersions
            The version as an ApiVersions enum.
        """
        if isinstance(version, ApiVersions):
            return version
        elif isinstance(version, str) and version.isnumeric():
            return ApiVersions(int(version))
        elif isinstance(version, int):
            return ApiVersions(version)
        else:
            raise ValueError("API version must be an integer, string or an ApiVersions enum.")



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the Prepare Tools service implementation for v0."""

import grpc

from ansys.geometry.core.errors import protect_grpc

from ..base.conversions import from_measurement_to_server_length
from ..base.prepare_tools import GRPCPrepareToolsService
from .conversions import (
    build_grpc_id,
    from_enclosure_options_to_grpc_enclosure_options,
    serialize_tracker_command_response,
)


class GRPCPrepareToolsServiceV0(GRPCPrepareToolsService):
    """Prepare tools service for gRPC communication with the Geometry server.

    This class provides methods to interact with the Geometry server's
    Prepare Tools service. It is specifically designed for the v0 version
    of the Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):  # noqa: D102
        from ansys.api.geometry.v0.preparetools_pb2_grpc import PrepareToolsStub

        self.stub = PrepareToolsStub(channel)

    @protect_grpc
    def extract_volume_from_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.preparetools_pb2 import ExtractVolumeFromFacesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = ExtractVolumeFromFacesRequest(
            sealing_faces=[build_grpc_id(face) for face in kwargs["sealing_faces"]],
            inside_faces=[build_grpc_id(face) for face in kwargs["inside_faces"]],
        )

        # Call the gRPC service
        response = self.stub.ExtractVolumeFromFaces(request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
            "created_bodies": [body.id for body in response.created_bodies],
        }

    @protect_grpc
    def extract_volume_from_edge_loops(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.preparetools_pb2 import ExtractVolumeFromEdgeLoopsRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = ExtractVolumeFromEdgeLoopsRequest(
            sealing_edges=[build_grpc_id(edge) for edge in kwargs["sealing_edges"]],
            inside_faces=[build_grpc_id(face) for face in kwargs["inside_faces"]],
        )

        # Call the gRPC service
        response = self.stub.ExtractVolumeFromEdgeLoops(request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
            "created_bodies": [body.id for body in response.created_bodies],
        }

    @protect_grpc
    def remove_rounds(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.models_pb2 import Face
        from ansys.api.geometry.v0.preparetools_pb2 import RemoveRoundsRequest
        from google.protobuf.wrappers_pb2 import BoolValue

        # Create the request - assumes all inputs are valid and of the proper type
        request = RemoveRoundsRequest(
            selection=[Face(id=round) for round in kwargs["rounds"]],
            auto_shrink=BoolValue(value=kwargs["auto_shrink"]),
        )

        # Call the gRPC service
        response = self.stub.RemoveRounds(request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.result,
        }

    @protect_grpc
    def share_topology(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.models_pb2 import Body
        from ansys.api.geometry.v0.preparetools_pb2 import ShareTopologyRequest
        from google.protobuf.wrappers_pb2 import BoolValue, DoubleValue

        # Create the request - assumes all inputs are valid and of the proper type
        request = ShareTopologyRequest(
            selection=[Body(id=body) for body in kwargs["bodies"]],
            tolerance=DoubleValue(value=from_measurement_to_server_length(kwargs["tolerance"])),
            preserve_instances=BoolValue(value=kwargs["preserve_instances"]),
        )

        # Call the gRPC service
        response = self.stub.ShareTopology(request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.result,
        }

    @protect_grpc
    def enhanced_share_topology(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.models_pb2 import Body
        from ansys.api.geometry.v0.preparetools_pb2 import ShareTopologyRequest
        from google.protobuf.wrappers_pb2 import BoolValue, DoubleValue

        # Create the request - assumes all inputs are valid and of the proper type
        request = ShareTopologyRequest(
            selection=[Body(id=body) for body in kwargs["bodies"]],
            tolerance=DoubleValue(value=from_measurement_to_server_length(kwargs["tolerance"])),
            preserve_instances=BoolValue(value=kwargs["preserve_instances"]),
        )

        # Call the gRPC service
        response = self.stub.EnhancedShareTopology(request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.success,
            "found": response.found,
            "repaired": response.repaired,
            "created_bodies_monikers": response.created_bodies_monikers,
            "modified_bodies_monikers": response.modified_bodies_monikers,
            "deleted_bodies_monikers": response.deleted_bodies_monikers,
        }

    @protect_grpc
    def find_logos(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.models_pb2 import FindLogoOptions
        from ansys.api.geometry.v0.preparetools_pb2 import FindLogosRequest

        # Check height objects
        min_height = (
            from_measurement_to_server_length(kwargs["min_height"])
            if kwargs["min_height"] is not None
            else None
        )
        max_height = (
            from_measurement_to_server_length(kwargs["max_height"])
            if kwargs["max_height"] is not None
            else None
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindLogosRequest(
            bodies=[build_grpc_id(body) for body in kwargs["bodies"]],
            options=FindLogoOptions(
                min_height=min_height,
                max_height=max_height,
            ),
        )

        # Call the gRPC service
        response = self.stub.FindLogos(request)

        # Return the response - formatted as a dictionary
        return {
            "id": response.id,
            "face_ids": [face.id for face in response.logo_faces],
        }

    @protect_grpc
    def find_and_remove_logos(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.models_pb2 import FindLogoOptions
        from ansys.api.geometry.v0.preparetools_pb2 import FindLogosRequest

        # Check height objects
        min_height = (
            from_measurement_to_server_length(kwargs["min_height"])
            if kwargs["min_height"] is not None
            else None
        )
        max_height = (
            from_measurement_to_server_length(kwargs["max_height"])
            if kwargs["max_height"] is not None
            else None
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = FindLogosRequest(
            bodies=[build_grpc_id(body) for body in kwargs["bodies"]],
            options=FindLogoOptions(
                min_height=min_height,
                max_height=max_height,
            ),
        )

        # Call the gRPC service
        response = self.stub.FindAndRemoveLogos(request)

        # Return the response - formatted as a dictionary
        return {"success": response.success}

    @protect_grpc
    def remove_logo(self, **kwargs):  # noqa: D102
        from ansys.api.geometry.v0.preparetools_pb2 import RemoveLogoRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = RemoveLogoRequest(
            face_ids=[build_grpc_id(id) for id in kwargs["face_ids"]],
        )

        # Call the gRPC service
        response = self.stub.RemoveLogo(request)

        # Return the response - formatted as a dictionary
        return {"success": response.success}

    @protect_grpc
    def detect_helixes(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.models_pb2 import DetectHelixesOptions
        from ansys.api.geometry.v0.preparetools_pb2 import DetectHelixesRequest

        from ansys.geometry.core.shapes.parameterization import Interval

        from ..base.conversions import (
            from_measurement_to_server_length,
            to_distance,
        )
        from .conversions import (
            from_grpc_curve_to_curve,
            from_grpc_point_to_point3d,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = DetectHelixesRequest(
            body_ids=[build_grpc_id(body) for body in kwargs["bodies"]],
            options=DetectHelixesOptions(
                min_radius=from_measurement_to_server_length(kwargs["min_radius"]),
                max_radius=from_measurement_to_server_length(kwargs["max_radius"]),
                fit_radius_error=from_measurement_to_server_length(kwargs["fit_radius_error"]),
            ),
        )

        # Call the gRPC service
        response = self.stub.DetectHelixes(request)

        # If no helixes, return empty dictionary
        if len(response.helixes) == 0:
            return {"helixes": []}

        # Return the response - formatted as a dictionary
        return {
            "helixes": [
                {
                    "trimmed_curve": {
                        "geometry": from_grpc_curve_to_curve(helix.trimmed_curve.curve),
                        "start": from_grpc_point_to_point3d(helix.trimmed_curve.start),
                        "end": from_grpc_point_to_point3d(helix.trimmed_curve.end),
                        "interval": Interval(
                            helix.trimmed_curve.interval_start, helix.trimmed_curve.interval_end
                        ),
                        "length": to_distance(helix.trimmed_curve.length).value,
                    },
                    "edges": [
                        {
                            "id": edge.id,
                            "parent_id": edge.parent.id,
                            "curve_type": edge.curve_type,
                            "is_reversed": edge.is_reversed,
                        }
                        for edge in helix.edges
                    ],
                }
                for helix in response.helixes
            ]
        }

    @protect_grpc
    def create_box_enclosure(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.models_pb2 import Body as GRPCBody
        from ansys.api.geometry.v0.preparetools_pb2 import CreateEnclosureBoxRequest

        grpc_enclosure_options = from_enclosure_options_to_grpc_enclosure_options(
            kwargs["enclosure_options"]
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateEnclosureBoxRequest(
            bodies=[GRPCBody(id=id) for id in kwargs["body_ids"]],
            x_low=from_measurement_to_server_length(kwargs["x_low"]),
            x_high=from_measurement_to_server_length(kwargs["x_high"]),
            y_low=from_measurement_to_server_length(kwargs["y_low"]),
            y_high=from_measurement_to_server_length(kwargs["y_high"]),
            z_low=from_measurement_to_server_length(kwargs["z_low"]),
            z_high=from_measurement_to_server_length(kwargs["z_high"]),
            enclosure_options=grpc_enclosure_options,
        )

        # Call the gRPC service
        response = self.stub.CreateEnclosureBox(request)

        # Return the response - formatted as a dictionary
        serialized_tracker_response = serialize_tracker_command_response(
            response=response.command_response
        )
        return {
            "success": response.success,
            "created_bodies": [body.id for body in response.created_bodies],
            "tracker_response": serialized_tracker_response,
        }

    @protect_grpc
    def create_cylinder_enclosure(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.models_pb2 import Body as GRPCBody
        from ansys.api.geometry.v0.preparetools_pb2 import CreateEnclosureCylinderRequest

        grpc_enclosure_options = from_enclosure_options_to_grpc_enclosure_options(
            kwargs["enclosure_options"]
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateEnclosureCylinderRequest(
            bodies=[GRPCBody(id=id) for id in kwargs["body_ids"]],
            axial_distance_low=from_measurement_to_server_length(kwargs["axial_distance_low"]),
            axial_distance_high=from_measurement_to_server_length(kwargs["axial_distance_high"]),
            radial_distance=from_measurement_to_server_length(kwargs["radial_distance"]),
            enclosure_options=grpc_enclosure_options,
        )

        # Call the gRPC service
        response = self.stub.CreateEnclosureCylinder(request)

        # Return the response - formatted as a dictionary
        serialized_tracker_response = serialize_tracker_command_response(
            response=response.command_response
        )
        return {
            "success": response.success,
            "created_bodies": [body.id for body in response.created_bodies],
            "tracker_response": serialized_tracker_response,
        }

    @protect_grpc
    def create_sphere_enclosure(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.models_pb2 import Body as GRPCBody
        from ansys.api.geometry.v0.preparetools_pb2 import CreateEnclosureSphereRequest

        grpc_enclosure_options = from_enclosure_options_to_grpc_enclosure_options(
            kwargs["enclosure_options"]
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateEnclosureSphereRequest(
            bodies=[GRPCBody(id=id) for id in kwargs["body_ids"]],
            radial_distance=from_measurement_to_server_length(kwargs["radial_distance"]),
            enclosure_options=grpc_enclosure_options,
        )

        # Call the gRPC service
        response = self.stub.CreateEnclosureSphere(request)

        # Return the response - formatted as a dictionary
        serialized_tracker_response = serialize_tracker_command_response(
            response=response.command_response
        )
        return {
            "success": response.success,
            "created_bodies": [body.id for body in response.created_bodies],
            "tracker_response": serialized_tracker_response,
        }

    @protect_grpc
    def detect_sweepable_bodies(self, **kwargs):  # noqa: D102
        from ansys.api.geometry.v0.preparetools_pb2 import DetectSweepableBodiesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = DetectSweepableBodiesRequest(
            bodies=[build_grpc_id(id=id) for id in kwargs["body_ids"]],
            get_source_target_faces=kwargs["get_source_target_faces"],
        )

        # Call the gRPC service
        response = self.stub.DetectSweepableBodies(request)

        # Return the response - formatted as a dictionary
        return {
            "results": [
                {"sweepable": result.result, "face_ids": [face.id for face in result.face_ids]}
                for result in response.response_data
            ]
        }



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides functions for performing common checks."""

from collections.abc import Iterable
import functools
from typing import TYPE_CHECKING, Type
import warnings

import numpy as np
from pint import Unit
import semver

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.design import Design
    from ansys.geometry.core.shapes.curves.trimmed_curve import TrimmedCurve
    from ansys.geometry.core.shapes.surfaces.trimmed_surface import TrimmedSurface
    from ansys.geometry.core.sketch.sketch import Sketch


def ensure_design_is_active(method):
    """Make sure that the design is active before executing a method.

    This function is necessary to be called whenever we do any operation
    on the design. If we are just accessing information of the class, it
    is not necessary to call this.
    """

    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        import ansys.geometry.core as pyansys_geometry
        from ansys.geometry.core.errors import GeometryRuntimeError

        if pyansys_geometry.DISABLE_ACTIVE_DESIGN_CHECK:
            # If the user has disabled the check, then we can skip it
            return method(self, *args, **kwargs)

        # Check if the current design is active... otherwise activate it
        def get_design_ref(obj) -> "Design":
            if hasattr(obj, "_modeler"):  # In case of a Design object
                return obj
            elif hasattr(
                obj, "_parent_component"
            ):  # In case of a Body, Component, DesignPoint, Beam
                # Recursive call
                return get_design_ref(obj._parent_component)
            elif hasattr(obj, "_body"):  # In case of a Face, Edge
                # Recursive call
                return get_design_ref(obj._body._parent_component)
            else:
                raise ValueError("Unable to find the design reference.")

        # Get the design reference
        design = get_design_ref(self)

        # Verify whether the Design has been closed on the backend
        if design.is_closed:
            raise GeometryRuntimeError(
                "The design has been closed on the backend. Cannot perform any operations on it."
            )

        # Finally, call method
        return method(self, *args, **kwargs)

    return wrapper


def check_is_float_int(param: object, param_name: str | None = None) -> None:
    """Check if a parameter has a float or integer value.

    Parameters
    ----------
    param : object
        Object instance to check.
    param_name : str, default: None
        Parameter name (if any).

    Raises
    ------
    TypeError
        If the parameter does not have a float or integer value.
    """
    if not isinstance(param, (int, float)):
        raise TypeError(
            "The parameter should have a float or integer value."
            if param_name is None
            else f"The parameter '{param_name}' should have a float or integer value."
        )


def check_ndarray_is_float_int(param: np.ndarray, param_name: str | None = None) -> None:
    """Check if a :class:`numpy.ndarray <numpy.ndarray>` has float/integer types.

    Parameters
    ----------
    param : ~numpy.ndarray
        :class:`numpy.ndarray <numpy.ndarray>` instance to check.
    param_name : str, default: None
        :class:`numpy.ndarray <numpy.ndarray>` instance name (if any).

    Raises
    ------
    TypeError
        If the :class:`numpy.ndarray <numpy.ndarray>` instance does not
        have float or integer values.
    """
    param_data = np.ravel(param)

    if not np.issubdtype(param.dtype, np.number) or not all(
        isinstance(data, (int, float)) for data in param_data.data
    ):
        raise TypeError(
            "The numpy.ndarray should contain float or integer values."
            if param_name is None
            else f"The numpy.ndarray '{param_name}' should contain float or integer values."
        )


def check_ndarray_is_not_none(param: np.ndarray, param_name: str | None = None) -> None:
    """Check if a :class:`numpy.ndarray <numpy.ndarray>` is all ``None``.

    Parameters
    ----------
    param : ~numpy.ndarray
        :class:`numpy.ndarray <numpy.ndarray>` instance to check.
    param_name : str, default: None
        :class:`numpy.ndarray <numpy.ndarray>` instance name (if any).

    Raises
    ------
    ValueError
        If the :class:`numpy.ndarray <numpy.ndarray>` instance has a value
        of ``None`` for all parameters.
    """
    param_data = np.ravel(param)
    if all(value is None for value in param_data):
        raise ValueError(
            "The numpy.ndarray should not have 'None' for all parameter values."
            if param_name is None
            else f"The numpy.ndarray '{param_name}' should not have 'None' for all parameter values."  # noqa: E501
        )


def check_ndarray_is_all_nan(param: np.ndarray, param_name: str | None = None) -> None:
    """Check if a :class:`numpy.ndarray <numpy.ndarray>` is all nan-valued.

    Parameters
    ----------
    param : ~numpy.ndarray
        :class:`numpy.ndarray <numpy.ndarray>` instance to check.
    param_name : str or None, default: None
        :class:`numpy.ndarray <numpy.ndarray>` instance name (if any).

    Raises
    ------
    ValueError
        If the :class:`numpy.ndarray <numpy.ndarray>` instance is all nan-valued.
    """
    if np.isnan(param).all():
        raise ValueError(
            "The numpy.ndarray should not be a nan numpy.ndarray."
            if param_name is None
            else f"The numpy.ndarray '{param_name}' should not be a nan numpy.ndarray."
        )


def check_ndarray_is_non_zero(param: np.ndarray, param_name: str | None = None) -> None:
    """Check if a :class:`numpy.ndarray <numpy.ndarray>` is zero-valued.

    Parameters
    ----------
    param : ~numpy.ndarray
        :class:`numpy.ndarray <numpy.ndarray>` instance to check.
    param_name : str, default: None
        :class:`numpy.ndarray <numpy.ndarray>` instance name (if any).

    Raises
    ------
    ValueError
        If the :class:`numpy.ndarray <numpy.ndarray>` instance is zero-valued.
    """
    param_data = np.ravel(param)
    if all(value == 0 for value in param_data):
        raise ValueError(
            "The numpy.ndarray should not be a numpy.ndarray of zeros."
            if param_name is None
            else f"The numpy.ndarray '{param_name}' should not be a numpy.ndarray of zeros."
        )


def check_pint_unit_compatibility(input: Unit, expected: Unit) -> None:
    """Check if input :class:`pint.Unit` is compatible with the expected input.

    Parameters
    ----------
    input : ~pint.Unit
        :class:`pint.Unit` input.
    expected : ~pint.Unit
        :class:`pint.Unit` expected dimensionality.

    Raises
    ------
    TypeError
        If the input is not compatible with the :class:`pint.Unit` class.
    """
    if not input.is_compatible_with(expected):
        raise TypeError(
            f"The pint.Unit provided as an input should be a {expected.dimensionality} quantity."
        )


def check_type_equivalence(input: object, expected: object) -> None:
    """Check if an input object is of the same class as an expected object.

    Parameters
    ----------
    input : object
        Input object.
    expected : object
        Expected object.

    Raises
    ------
    TypeError
        If the objects are not of the same class.
    """
    if not isinstance(input, type(expected)):
        raise TypeError(
            f"Provided type {type(input)} is invalid. Type {type(expected)} is expected."
        )


def check_type(input: object, expected_type: Type | tuple[Type, ...]) -> None:
    """Check if an input object is of the same type as expected types.

    Parameters
    ----------
    input : object
        Input object.
    expected_type : Type | tuple[Type, ...]
        One or more types to compare the input object against.

    Raises
    ------
    TypeError
        If the object does not match the one or more expected types.
    """
    if not isinstance(input, expected_type):
        raise TypeError(
            f"Provided type {type(input)} is invalid. Type {expected_type} is expected."
        )


def check_type_all_elements_in_iterable(
    input: Iterable, expected_type: Type | tuple[Type, ...]
) -> None:
    """Check if all elements in an iterable are of the same type as expected.

    Parameters
    ----------
    input : Iterable
        Input iterable.
    expected_type : Type | tuple[Type, ...]
        One or more types to compare the input object against.

    Raises
    ------
    TypeError
        If one of the elements in the iterable does not match the one or more expected types.
    """
    for elem in input:
        check_type(elem, expected_type)


def check_nurbs_compatibility(
    backend_version: semver.Version,
    sketch: "Sketch" = None,
    curves: list["TrimmedCurve"] = None,
    surfaces: list["TrimmedSurface"] = None,
) -> None:
    """Check if the inputs require NURBS functionality and it is available.

    Parameters
    ----------
    backend_version : semver.Version
        Backend version to check against.
    sketch : Sketch, default: None
        Sketch to check for NURBS geometry.
    curves : list[TrimmedCurve], default: None
        List of TrimmedCurve to check for NURBS geometry.
    surfaces : list[TrimmedSurface], default: None
        List of TrimmedSurface to check for NURBS geometry.

    Raises
    ------
    GeometryRuntimeError
        If inputs contain NURBS functionality but the backend is older than 26R1.
    """
    from ansys.geometry.core.errors import GeometryRuntimeError
    from ansys.geometry.core.shapes.curves.nurbs import NURBSCurve
    from ansys.geometry.core.shapes.surfaces.nurbs import NURBSSurface
    from ansys.geometry.core.sketch.nurbs import SketchNurbs

    requires_nurbs = False

    if sketch is not None:
        if any(isinstance(edge, SketchNurbs) for edge in sketch.edges):
            requires_nurbs = True

    if curves is not None:
        if any(isinstance(curve.geometry, NURBSCurve) for curve in curves):
            requires_nurbs = True

    if surfaces is not None:
        if any(isinstance(surface.geometry, NURBSSurface) for surface in surfaces):
            requires_nurbs = True

    if requires_nurbs:
        min_version = semver.Version(26, 1, 0)
        comp = min_version.compare(backend_version)
        if comp == 1:
            raise GeometryRuntimeError(
                "NURBS functionality requires a minimum Ansys release version of 26R1, "
                + f"but the current version used is {backend_version}."
            )


def min_backend_version(major: int, minor: int, service_pack: int):
    """Compare a minimum required version to the current backend version.

    Parameters
    ----------
    major : int
        Minimum major version required by the method.
    minor : int
        Minimum minor version required by the method.
    service_pack : int
        Minimum service pack version required by the method.

    Raises
    ------
    GeometryRuntimeError
        If the method version is higher than the backend version.
    GeometryRuntimeError
        If the client is not available.
    """
    # Lazy import to avoid circular imports
    from ansys.geometry.core.errors import GeometryRuntimeError
    from ansys.geometry.core.logger import LOG

    def backend_version_decorator(method):
        @functools.wraps(method)
        def wrapper(self, *args, **kwargs):
            method_version = semver.Version(major, minor, service_pack)
            if hasattr(self, "_grpc_client"):
                if self._grpc_client is None:
                    raise GeometryRuntimeError(
                        "The client is not available. You must initialize the client first."
                    )
                elif self._grpc_client.backend_version is not None:
                    comp = method_version.compare(self._grpc_client.backend_version)
                    # if comp is 1, method version is higher than backend version.
                    if comp == 1:
                        # Check if the version is "0.0.0" (i.e., the version is not available)
                        if str(self._grpc_client.backend_version) == "0.0.0":
                            raise GeometryRuntimeError(
                                f"The method '{method.__name__}' requires a minimum Ansys release version of "  # noqa: E501
                                + f"{method_version}, but the current version used is 24.1.0 or lower."  # noqa: E501
                            )
                        else:
                            raise GeometryRuntimeError(
                                f"The method '{method.__name__}' requires a minimum Ansys release version of "  # noqa: E501
                                + f"{method_version}, but the current version used is "
                                + f"{self._grpc_client.backend_version}."
                            )
                    else:
                        return method(self, *args, **kwargs)
            else:
                LOG.warning("This object does not have a connection with the backend.")

        return wrapper

    return backend_version_decorator


def deprecated_method(
    alternative: str | None = None,
    info: str | None = None,
    version: str | None = None,
    remove: str | None = None,
):
    """Decorate a method as deprecated.

    Parameters
    ----------
    alternative : str, default: None
        Alternative method to use. If provided, the warning message will
        include the alternative method.
    info : str, default: None
        Additional information to include in the warning message.
    version : str, default: None
        Version where the method was deprecated.
    remove : str, default: None
        Version where the method will be removed.
    """

    def deprecated_decorator(method):
        @functools.wraps(method)
        def wrapper(*args, **kwargs):
            msg = f"The method '{method.__name__}' is deprecated."
            if alternative:
                msg += f" Use '{alternative}' instead."
            if info:
                msg += f" {info}"
            if version:
                msg += f" This method was deprecated in version {version}."
            if remove:
                msg += f" This method will be removed in version {remove}."
            warnings.warn(msg, DeprecationWarning)
            return method(*args, **kwargs)

        return wrapper

    return deprecated_decorator


def deprecated_argument(
    arg: str,
    alternative: str | None = None,
    info: str | None = None,
    version: str | None = None,
    remove: str | None = None,
):
    """Decorate a method argument as deprecated.

    Parameters
    ----------
    arg : str
        Argument to mark as deprecated.
    alternative : str, default: None
        Alternative argument to use. If provided, the warning message will
        include the alternative argument.
    info : str, default: None
        Additional information to include in the warning message.
    version : str, default: None
        Version where the method was deprecated.
    remove : str, default: None
        Version where the method will be removed.
    """

    def deprecated_decorator(method):
        @functools.wraps(method)
        def wrapper(*args, **kwargs):
            if arg in kwargs and kwargs[arg] is not None:
                msg = f"The argument '{arg}' in '{method.__name__}' is deprecated."
                if alternative:
                    msg += f" Use '{alternative}' instead."
                if info:
                    msg += f" {info}"
                if version:
                    msg += f" This argument was deprecated in version {version}."
                if remove:
                    msg += f" This argument will be removed in version {remove}."
                warnings.warn(msg, DeprecationWarning)

            return method(*args, **kwargs)

        return wrapper

    return deprecated_decorator


ERROR_GRAPHICS_REQUIRED = (
    "Graphics are required for this method. Please install the ``graphics`` target "
    " to use this method. You can install it by running `pip install ansys-geometry-core[graphics]`"
    " or `pip install ansys-geometry-core[all]`."
)
"""Message to display when graphics are required for a method."""

__GRAPHICS_AVAILABLE = None
"""Global variable to store the result of the graphics imports."""


def run_if_graphics_required():
    """Check if graphics are available."""
    global __GRAPHICS_AVAILABLE
    if __GRAPHICS_AVAILABLE is None:
        try:
            # Attempt to perform the imports
            import ansys.tools.visualization_interface  # noqa: F401
            import pygltflib  # noqa: F401
            import pyvista  # noqa: F401
            import trame  # noqa: F401
            import vtk  # noqa: F401

            __GRAPHICS_AVAILABLE = True
        except (ModuleNotFoundError, ImportError):
            __GRAPHICS_AVAILABLE = False

    if __GRAPHICS_AVAILABLE is False:
        raise ImportError(ERROR_GRAPHICS_REQUIRED)


def graphics_required(method):
    """Decorate a method as requiring graphics.

    Parameters
    ----------
    method : callable
        Method to decorate.

    Returns
    -------
    callable
        Decorated method.
    """

    @functools.wraps(method)
    def wrapper(*args, **kwargs):
        run_if_graphics_required()
        return method(*args, **kwargs)

    return wrapper


def kwargs_passed_not_accepted(method):
    """Check that no unexpected kwargs are passed to the method.

    This decorator will raise a TypeError if any keyword arguments are passed
    to the decorated method that don't correspond to the method's parameters.
    If the method has ``**kwargs`` in its signature, this decorator will raise an
    error for any kwargs passed to it (that are not explicitly accepted).

    Parameters
    ----------
    method : callable
        The method to decorate.

    Returns
    -------
    callable
        Decorated method that raises TypeError if unexpected kwargs are passed.

    Raises
    ------
    TypeError
        If unexpected keyword arguments are passed to the decorated method.

    Examples
    --------
    >>> @kwargs_passed_not_accepted
    ... def my_method(arg1, arg2):
    ...     return arg1 + arg2
    >>> my_method(1, 2)  # Works fine
    3
    >>> my_method(arg1=1, arg2=2)  # Works fine
    3
    >>> my_method(1, 2, invalid_arg=3)  # Raises TypeError
    TypeError: The following keyword arguments are not accepted
    in the method 'my_method': invalid_arg.
    >>> @kwargs_passed_not_accepted
    ... def my_method_with_kwargs(arg1, arg2, **kwargs):
    ...     return arg1 + arg2
    >>> my_method_with_kwargs(1, 2, invalid_arg=3)  # Raises TypeError
    TypeError: The following keyword arguments are not accepted
    in the method 'my_method_with_kwargs': invalid_arg.
    """
    import inspect

    @functools.wraps(method)
    def wrapper(*args, **kwargs):
        # Get the method signature
        sig = inspect.signature(method)

        # Check if method has **kwargs parameter
        has_var_keyword = any(
            param.kind == inspect.Parameter.VAR_KEYWORD for param in sig.parameters.values()
        )

        # If method has **kwargs and kwargs are passed...
        if has_var_keyword and len(kwargs) > 0:
            # Retrieve a list of all parameter names excluding **kwargs
            param_names = {
                name
                for name, param in sig.parameters.items()
                if param.kind != inspect.Parameter.VAR_KEYWORD
            }

            # Identify unexpected kwargs
            unexpected_kwargs = [key for key in kwargs.keys() if key not in param_names]

            if unexpected_kwargs:
                raise TypeError(
                    "The following keyword arguments are not accepted in the"
                    + f" method '{method.__name__}': {', '.join(unexpected_kwargs)}."
                )

        return method(*args, **kwargs)

    return wrapper



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Utility functions for plotting and visualization."""

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:  # pragma: no cover
    import pyvista as pv

    from ansys.geometry.core.math.point import Point3D
    from ansys.geometry.core.math.vector import Vector3D
    from ansys.geometry.core.typing import Real


def create_elliptical_polydata(
    origin: "Point3D",
    dir_x: "Vector3D",
    dir_y: "Vector3D",
    dir_z: "Vector3D",
    major_radius: "Real",
    minor_radius: "Real | None" = None,
    num_points: int = 100,
) -> "pv.PolyData":
    """Create PyVista PolyData for an ellipse or circle.

    This function creates a discretized representation of an ellipse (or circle)
    as a closed loop of line segments for visualization purposes.

    Parameters
    ----------
    origin : Point3D
        Center point of the ellipse/circle.
    dir_x : Vector3D
        Direction vector for the major axis (or X axis for circles).
    dir_y : Vector3D
        Direction vector for the minor axis (or Y axis for circles).
    dir_z : Vector3D
        Normal direction vector.
    major_radius : Real
        Major radius of the ellipse. For circles, this is the radius.
    minor_radius : Real, optional
        Minor radius of the ellipse. If None, defaults to major_radius
        (creating a circle).
    num_points : int, default: 100
        Number of points to use for discretization.

    Returns
    -------
    pyvista.PolyData
        VTK PolyData representation with line segments forming a closed loop.

    Notes
    -----
    For a circle, either pass the same value for both radii or only provide
    major_radius (minor_radius will default to major_radius).
    """
    import pyvista as pv

    # Default minor_radius to major_radius for circles
    if minor_radius is None:
        minor_radius = major_radius

    # Create parametric points
    # Use endpoint=False to avoid duplicating the first/last point in the closed loop
    theta = np.linspace(0, 2 * np.pi, num_points, endpoint=False)

    # Calculate points using ellipse parametric equations
    points = np.column_stack(
        [
            (
                origin[0]
                + major_radius * np.cos(theta) * dir_x[0]
                + minor_radius * np.sin(theta) * dir_y[0]
            ),
            (
                origin[1]
                + major_radius * np.cos(theta) * dir_x[1]
                + minor_radius * np.sin(theta) * dir_y[1]
            ),
            (
                origin[2]
                + major_radius * np.cos(theta) * dir_x[2]
                + minor_radius * np.sin(theta) * dir_y[2]
            ),
        ]
    )

    # Create line segments forming a closed loop
    lines = np.column_stack(
        [np.full(num_points, 2), np.arange(num_points), np.roll(np.arange(num_points), -1)]
    ).ravel()

    return pv.PolyData(points, lines=lines)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the beams service implementation for v0."""

from ansys.api.geometry.v0.commands_pb2 import CreateBeamCircularProfileRequest
import grpc

from ansys.geometry.core.errors import protect_grpc

from ..base.beams import GRPCBeamsService
from ..base.conversions import from_measurement_to_server_length, to_distance
from .conversions import (
    build_grpc_id,
    from_grpc_curve_to_curve,
    from_grpc_frame_to_frame,
    from_grpc_material_to_material,
    from_grpc_point_to_point3d,
    from_plane_to_grpc_plane,
    from_point3d_to_grpc_point,
)


class GRPCBeamsServiceV0(GRPCBeamsService):
    """Beams service for gRPC communication with the Geometry server.

    This class provides methods to interact with the Geometry server's
    beams service. It is specifically designed for the v0 version of the
    Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):  # noqa: D102
        from ansys.api.geometry.v0.commands_pb2_grpc import CommandsStub

        self.commands_stub = CommandsStub(channel)

    @protect_grpc
    def create_beam_segments(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.commands_pb2 import CreateBeamSegmentsRequest
        from ansys.api.geometry.v0.models_pb2 import Line

        # Create the gRPC Line objects
        lines = []
        for segment in kwargs["segments"]:
            lines.append(
                Line(
                    start=from_point3d_to_grpc_point(segment[0]),
                    end=from_point3d_to_grpc_point(segment[1]),
                )
            )

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateBeamSegmentsRequest(
            profile=kwargs["profile_id"],
            parent=kwargs["parent_id"],
            lines=lines,
        )

        # Call the gRPC service
        resp = self.commands_stub.CreateBeamSegments(request)

        # Return the response - formatted as a dictionary
        return {
            "beam_ids": resp.ids,
        }

    @protect_grpc
    def create_descriptive_beam_segments(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.geometry.v0.commands_pb2 import CreateBeamSegmentsRequest
        from ansys.api.geometry.v0.models_pb2 import Line

        from ansys.geometry.core.shapes.parameterization import Interval, ParamUV

        # Create the gRPC Line objects
        lines = []
        for segment in kwargs["segments"]:
            lines.append(
                Line(
                    start=from_point3d_to_grpc_point(segment[0]),
                    end=from_point3d_to_grpc_point(segment[1]),
                )
            )

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateBeamSegmentsRequest(
            profile=kwargs["profile_id"],
            parent=kwargs["parent_id"],
            lines=lines,
        )

        # Call the gRPC service
        resp = self.commands_stub.CreateDescriptiveBeamSegments(request)

        # Return the response - formatted as a dictionary
        return {
            "created_beams": [
                {
                    "cross_section": {
                        "section_anchor": beam.cross_section.section_anchor,
                        "section_angle": beam.cross_section.section_angle,
                        "section_frame": from_grpc_frame_to_frame(beam.cross_section.section_frame),
                        "section_profile": [
                            [
                                {
                                    "geometry": from_grpc_curve_to_curve(curve.curve),
                                    "start": from_grpc_point_to_point3d(curve.start),
                                    "end": from_grpc_point_to_point3d(curve.end),
                                    "interval": Interval(curve.interval_start, curve.interval_end),
                                    "length": to_distance(curve.length).value,
                                }
                                for curve in curve_list.curves
                            ]
                            for curve_list in beam.cross_section.section_profile
                        ],
                    },
                    "properties": {
                        "area": beam.properties.area,
                        "centroid": ParamUV(beam.properties.centroid_x, beam.properties.centroid_y),
                        "warping_constant": beam.properties.warping_constant,
                        "ixx": beam.properties.ixx,
                        "ixy": beam.properties.ixy,
                        "iyy": beam.properties.iyy,
                        "shear_center": ParamUV(
                            beam.properties.shear_center_x, beam.properties.shear_center_y
                        ),
                        "torsional_constant": beam.properties.torsional_constant,
                    },
                    "id": beam.id.id,
                    "start": from_grpc_point_to_point3d(beam.shape.start),
                    "end": from_grpc_point_to_point3d(beam.shape.end),
                    "name": beam.name,
                    "is_deleted": beam.is_deleted,
                    "is_reversed": beam.is_reversed,
                    "is_rigid": beam.is_rigid,
                    "material": from_grpc_material_to_material(beam.material),
                    "shape": {
                        "geometry": from_grpc_curve_to_curve(beam.shape.curve),
                        "start": from_grpc_point_to_point3d(beam.shape.start),
                        "end": from_grpc_point_to_point3d(beam.shape.end),
                        "interval": Interval(beam.shape.interval_start, beam.shape.interval_end),
                        "length": to_distance(beam.shape.length).value,
                    },
                    "beam_type": beam.type,
                }
                for beam in resp.created_beams
            ],
        }

    @protect_grpc
    def delete_beam(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(kwargs["beam_id"])

        # Call the gRPC service
        _ = self.commands_stub.DeleteBeam(request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def delete_beam_profile(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = build_grpc_id(id=kwargs["id"])

        # Call the gRPC service
        _ = self.commands_stub.DeleteBeamProfile(request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def create_beam_circular_profile(self, **kwargs) -> dict:  # noqa: D102
        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateBeamCircularProfileRequest(
            origin=from_point3d_to_grpc_point(kwargs["center"]),
            radius=from_measurement_to_server_length(kwargs["radius"]),
            plane=from_plane_to_grpc_plane(kwargs["plane"]),
            name=kwargs["name"],
        )

        # Call the gRPC service
        response = self.commands_stub.CreateBeamCircularProfile(request)

        # Return the response - formatted as a dictionary
        return {"id": response.id}



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for managing a bounding box."""

from typing import Union

from beartype import beartype as check_input_types

from ansys.geometry.core.math.misc import intersect_interval
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.misc.accuracy import Accuracy
from ansys.geometry.core.misc.measurements import DEFAULT_UNITS
from ansys.geometry.core.typing import Real


class BoundingBox:
    """Maintains the box structure for Bounding Boxes.

    Parameters
    ----------
    min_corner : Point3D
        Minimum corner for the box.
    max_corner : Point3D
        Maximum corner for the box.
    center : Point3D
        Center of the box.
    """

    @check_input_types
    def __init__(self, min_corner: Point3D, max_corner: Point3D, center: Point3D = None):
        """Initialize the ``BoundingBox`` class."""
        self._min_corner = min_corner
        self._max_corner = max_corner

        if center is not None:
            self._center = center
        else:
            mid_x = (max_corner.x.m - min_corner.x.m) / 2
            mid_y = (max_corner.y.m - min_corner.y.m) / 2
            mid_z = (max_corner.z.m - min_corner.z.m) / 2
            self._center = Point3D([mid_x, mid_y, mid_z])

    @property
    def min_corner(self) -> Point3D:
        """Minimum corner of the bounding box."""
        return self._min_corner

    @property
    def max_corner(self) -> Point3D:
        """Maximum corner of the bounding box."""
        return self._max_corner

    @property
    def center(self) -> Point3D:
        """Center of the bounding box."""
        return self._center

    @check_input_types
    def contains_point(self, point: Point3D) -> bool:
        """Evaluate whether a point lies within the box.

        Parameters
        ----------
        point : Point3D
            Point to compare against the bounds.

        Returns
        -------
        bool
            ``True`` if the point is contained in the bounding box. Otherwise, ``False``.
        """
        return self.contains_point_components(
            point.x.m_as(DEFAULT_UNITS.LENGTH),
            point.y.m_as(DEFAULT_UNITS.LENGTH),
            point.z.m_as(DEFAULT_UNITS.LENGTH),
        )

    @check_input_types
    def contains_point_components(self, x: Real, y: Real, z: Real) -> bool:
        """Check if point components are within box.

        Parameters
        ----------
        x : Real
            Point X component to compare against the bounds.
        y : Real
            Point Y component to compare against the bounds.
        z : Real
            Point Z component to compare against the bounds.

        Returns
        -------
        bool
            ``True`` if the components are contained in the bounding box. Otherwise, ``False``.
        """
        return (
            Accuracy.length_is_greater_than_or_equal(x, self._min_corner.x.m)
            and Accuracy.length_is_greater_than_or_equal(y, self._min_corner.y.m)
            and Accuracy.length_is_greater_than_or_equal(z, self._min_corner.z.m)
            and Accuracy.length_is_less_than_or_equal(x, self._max_corner.x.m)
            and Accuracy.length_is_less_than_or_equal(y, self._max_corner.y.m)
            and Accuracy.length_is_less_than_or_equal(z, self._max_corner.z.m)
        )

    @check_input_types
    def __eq__(self, other: "BoundingBox") -> bool:
        """Equals operator for the ``BoundingBox`` class."""
        return (
            self._min_corner.x == other._min_corner.x
            and self._max_corner.x == other._max_corner.x
            and self._min_corner.y == other._min_corner.y
            and self._max_corner.y == other._max_corner.y
            and self._min_corner.z == other._min_corner.z
            and self._max_corner.z == other._max_corner.z
        )

    @check_input_types
    def __ne__(self, other: "BoundingBox") -> bool:
        """Not equals operator for the ``BoundingBox`` class."""
        return not self == other

    @staticmethod
    def intersect_bboxes(box_1: "BoundingBox", box_2: "BoundingBox") -> Union[None, "BoundingBox"]:
        """Find the intersection of 2 BoundingBox objects.

        Parameters
        ----------
        box_1: BoundingBox
            The box to consider the intersection of with respect to box_2.
        box_2: BoundingBox
            The box to consider the intersection of with respect to box_1.

        Returns
        -------
        BoundingBox:
            The box representing the intersection of the two passed in boxes.
        """
        intersect, min_x, max_x = intersect_interval(
            box_1._min_corner.x.m,
            box_2._min_corner.x.m,
            box_1._max_corner.x.m,
            box_2._max_corner.x.m,
        )
        if not intersect:
            return None

        intersect, min_y, max_y = intersect_interval(
            box_1._min_corner.y.m,
            box_2._min_corner.y.m,
            box_1._max_corner.y.m,
            box_2._max_corner.y.m,
        )
        if not intersect:
            return None

        intersect, min_z, max_z = intersect_interval(
            box_1._min_corner.z.m,
            box_2._min_corner.z.m,
            box_1._max_corner.z.m,
            box_2._max_corner.z.m,
        )
        if not intersect:
            return None

        min_point = Point3D([min_x, min_y, min_z])
        max_point = Point3D([max_x, max_y, max_z])
        return BoundingBox(min_point, max_point)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides for creating and managing a beam."""

from enum import Enum
from typing import TYPE_CHECKING

from beartype import beartype as check_input_types

from ansys.geometry.core.materials.material import Material
from ansys.geometry.core.math.frame import Frame
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.math.vector import UnitVector3D
from ansys.geometry.core.misc.auxiliary import get_design_from_component
from ansys.geometry.core.misc.checks import check_type
from ansys.geometry.core.misc.measurements import Distance
from ansys.geometry.core.shapes.curves.trimmed_curve import TrimmedCurve
from ansys.geometry.core.shapes.parameterization import ParamUV

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.component import Component
    from ansys.geometry.core.designer.selection import NamedSelection


class BeamType(Enum):
    """Provides values for the beam types supported."""

    BEAM = 0
    SPRING = 1
    LINK_TRUSS = 2
    CABLE = 3
    PIPE = 4
    THERMALFLUID = 5
    UNKNOWN = 6


class SectionAnchorType(Enum):
    """Provides values for the section anchor types supported."""

    CENTROID = 0
    SHEAR_CENTER = 1
    ANCHOR_LOCATION = 2


class BeamProfile:
    """Represents a single beam profile organized within the design assembly.

    This profile synchronizes to a design within a supporting Geometry service instance.

    Parameters
    ----------
    id : str
        Server-defined ID for the beam profile.
    name : str
        User-defined label for the beam profile.

    Notes
    -----
    ``BeamProfile`` objects are expected to be created from the ``Design`` object.
    This means that you are not expected to instantiate your own ``BeamProfile``
    object. You should call the specific ``Design`` API for the ``BeamProfile`` desired.
    """

    def __init__(self, id: str, name: str):
        """Initialize ``BeamProfile`` class."""
        self._id = id
        self._name = name

    @property
    def id(self) -> str:
        """ID of the beam profile."""
        return self._id

    @property
    def name(self) -> str:
        """Name of the beam profile."""
        return self._name


class BeamCircularProfile(BeamProfile):
    """Represents a single circular beam profile.

    This profile synchronizes to a design within a supporting Geometry service instance.

    Parameters
    ----------
    id : str
        Server-defined ID for the beam profile.
    name : str
        User-defined label for the beam profile.
    radius : Distance
        Radius of the circle.
    center: Point3D
        3D point representing the center of the circle.
    direction_x: UnitVector3D
        X-axis direction.
    direction_y: UnitVector3D
        Y-axis direction.

    Notes
    -----
    ``BeamProfile`` objects are expected to be created from the ``Design`` object.
    This means that you are not expected to instantiate your own ``BeamProfile``
    object. You should call the specific ``Design`` API for the ``BeamProfile`` desired.
    """

    def __init__(
        self,
        id: str,
        name: str,
        radius: Distance,
        center: Point3D,
        direction_x: UnitVector3D,
        direction_y: UnitVector3D,
    ):
        """Initialize ``BeamCircularProfile`` class."""
        super().__init__(id, name)

        # Store specific BeamCircularProfile variables
        self._radius = radius
        self._center = center
        self._dir_x = direction_x
        self._dir_y = direction_y

    @property
    def radius(self) -> Distance:
        """Radius of the circular beam profile."""
        return self._radius

    @property
    def center(self) -> Point3D:
        """Center of the circular beam profile."""
        return self._center

    @property
    def direction_x(self) -> UnitVector3D:
        """X-axis direction of the circular beam profile."""
        return self._dir_x

    @property
    def direction_y(self) -> UnitVector3D:
        """Y-axis direction of the circular beam profile."""
        return self._dir_y

    def __repr__(self) -> str:
        """Represent the ``BeamCircularProfile`` as a string."""
        lines = [f"ansys.geometry.core.designer.BeamCircularProfile {hex(id(self))}"]
        lines.append(f"  Name                 : {self.name}")
        lines.append(f"  Radius               : {str(self.radius.value)}")
        lines.append(
            f"  Center               : [{','.join([str(x) for x in self.center])}] in meters"
        )
        lines.append(f"  Direction x          : [{','.join([str(x) for x in self.direction_x])}]")
        lines.append(f"  Direction y          : [{','.join([str(x) for x in self.direction_y])}]")
        return "\n".join(lines)


class BeamCrossSectionInfo:
    """Represents the cross-section information for a beam.

    Parameters
    ----------
    section_anchor : SectionAnchorType
        Specifies how the beam section is anchored to the beam path.
    section_angle : float
        The rotation angle of the cross section clockwise from the default perpendicular of the
        beam path.
    section_frame : Frame
        The section frame at the start of the beam.
    section_profile : BeamProfile
        The section profile in the XY plane.
    """

    @check_input_types
    def __init__(
        self,
        section_anchor: SectionAnchorType,
        section_angle: float,
        section_frame: Frame,
        section_profile: list[list[TrimmedCurve]] | None,
    ):
        """Initialize ``BeamCrossSectionInfo`` class."""
        self._section_anchor = section_anchor
        self._section_angle = section_angle
        self._section_frame = section_frame
        self._section_profile = section_profile

    @property
    def section_anchor(self) -> SectionAnchorType:
        """Specifies how the beam section is anchored to the beam path."""
        return self._section_anchor

    @property
    def section_angle(self) -> float:
        """Rotation angle of the cross section clockwise from the perpendicular of the beam path."""
        return self._section_angle

    @property
    def section_frame(self) -> Frame:
        """The section frame at the start of the beam."""
        return self._section_frame

    @property
    def section_profile(self) -> list[list[TrimmedCurve]] | None:
        """The section profile in the XY plane."""
        return self._section_profile

    def __repr__(self) -> str:
        """Represent the ``BeamCrossSectionInfo`` as a string."""
        lines = [f"ansys.geometry.core.designer.BeamCrossSectionInfo {hex(id(self))}"]
        lines.append(f"  Section Anchor       : {self.section_anchor.name}")
        lines.append(f"  Section Angle        : {self.section_angle}")
        lines.append(f"  Section Frame        : {self.section_frame}")
        lines.extend(
            ["\n", "  Section Profile info", "  -------------------", str(self.section_profile)]
        )
        return "\n".join(lines)


class BeamProperties:
    """Represents the properties of a beam.

    Parameters
    ----------
    area : float
        The cross-sectional area of the beam.
    centroid : ParamUV
        The centroid of the beam section.
    warping_constant : float
        The warping constant of the beam.
    ixx : float
        The moment of inertia about the x-axis.
    ixy : float
        The product of inertia.
    iyy : float
        The moment of inertia about the y-axis.
    shear_center : ParamUV
        The shear center of the beam.
    torsion_constant : float
        The torsion constant of the beam.
    """

    def __init__(
        self,
        area: float,
        centroid: ParamUV,
        warping_constant: float,
        ixx: float,
        ixy: float,
        iyy: float,
        shear_center: ParamUV,
        torsion_constant: float,
    ):
        """Initialize ``BeamProperties`` class."""
        check_type(area, float)
        check_type(centroid, ParamUV)
        check_type(warping_constant, float)
        check_type(ixx, float)
        check_type(ixy, float)
        check_type(iyy, float)
        check_type(shear_center, ParamUV)
        check_type(torsion_constant, float)

        self._area = area
        self._centroid = centroid
        self._warping_constant = warping_constant
        self._ixx = ixx
        self._ixy = ixy
        self._iyy = iyy
        self._shear_center = shear_center
        self._torsion_constant = torsion_constant

    @property
    def area(self) -> float:
        """The cross-sectional area of the beam."""
        return self._area

    @property
    def centroid(self) -> ParamUV:
        """The centroid of the beam section."""
        return self._centroid

    @property
    def warping_constant(self) -> float:
        """The warping constant of the beam."""
        return self._warping_constant

    @property
    def ixx(self) -> float:
        """The moment of inertia about the x-axis."""
        return self._ixx

    @property
    def ixy(self) -> float:
        """The product of inertia."""
        return self._ixy

    @property
    def iyy(self) -> float:
        """The moment of inertia about the y-axis."""
        return self._iyy

    @property
    def shear_center(self) -> ParamUV:
        """The shear center of the beam."""
        return self._shear_center

    @property
    def torsion_constant(self) -> float:
        """The torsion constant of the beam."""
        return self._torsion_constant


class Beam:
    """Represents a simplified solid body with an assigned 2D cross-section.

    This body synchronizes to a design within a supporting Geometry service instance.

    Parameters
    ----------
    id : str
        Server-defined ID for the body.
    name : str
        User-defined label for the body.
    start : Point3D
        Start of the beam line segment.
    end : Point3D
        End of the beam line segment.
    profile : BeamProfile
        Beam profile to use to create the beam.
    parent_component : Component
        Parent component to nest the new beam under within the design assembly.
    """

    def __init__(
        self,
        id: str,
        start: Point3D,
        end: Point3D,
        profile: BeamProfile | None,
        # TODO: Beams need BeamProfiles imported from existing design
        # https://github.com/ansys/pyansys-geometry/issues/1825
        parent_component: "Component",
        name: str = None,
        is_deleted: bool = False,
        is_reversed: bool = False,
        is_rigid: bool = False,
        material: Material = None,
        cross_section: BeamCrossSectionInfo = None,
        properties: BeamProperties = None,
        shape: TrimmedCurve = None,
        beam_type: BeamType = None,
    ):
        """Initialize ``Beam`` class."""
        from ansys.geometry.core.designer.component import Component

        check_type(id, str)
        check_type(start, Point3D)
        check_type(end, Point3D)
        check_type(profile, (type(None), BeamProfile))
        check_type(parent_component, Component)

        self._id = id
        self._start = start
        self._end = end
        self._profile = profile
        self._parent_component = parent_component
        self._is_alive = True
        self._name = name
        self._is_deleted = is_deleted
        self._is_reversed = is_reversed
        self._is_rigid = is_rigid
        self._material = material
        self._cross_section = cross_section
        self._properties = properties
        self._type = beam_type

    @property
    def id(self) -> str:
        """Service-defined ID of the beam."""
        return self._id

    @property
    def start(self) -> Point3D:
        """Start of the beam line segment."""
        return self._start

    @property
    def end(self) -> Point3D:
        """End of the beam line segment."""
        return self._end

    @property
    def profile(self) -> BeamProfile:
        """Beam profile of the beam line segment."""
        return self._profile

    @property
    def parent_component(self) -> "Component":
        """Component node that the beam is under."""
        return self._parent_component

    @property
    def is_alive(self) -> bool:
        """Flag indicating whether the beam is still alive on the server."""
        return self._is_alive

    def get_named_selections(self) -> list["NamedSelection"]:
        """Get the named selections that include this beam.

        Returns
        -------
        list[NamedSelection]
            List of named selections that include this beam.
        """
        included_ns = []
        for ns in get_design_from_component(self.parent_component).named_selections:
            if any(beam.id == self.id for beam in ns.beams):
                included_ns.append(ns)

        return included_ns

    def __repr__(self) -> str:
        """Represent the beam as a string."""
        lines = [f"ansys.geometry.core.designer.Beam {hex(id(self))}"]
        lines.append(f"  Exists               : {self.is_alive}")
        lines.append(
            f"  Start                : [{','.join([str(x) for x in self.start])}] in meters"
        )
        lines.append(f"  End                  : [{','.join([str(x) for x in self.end])}] in meters")
        lines.append(f"  Parent component     : {self.parent_component.name}")
        lines.extend(["\n", "  Beam Profile info", "  -----------------", str(self.profile)])
        return "\n".join(lines)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the bodies service implementation for v1."""

import grpc

from ansys.geometry.core.errors import protect_grpc
from ansys.geometry.core.misc.measurements import Distance

from ..base.bodies import GRPCBodyService
from ..base.conversions import from_measurement_to_server_length
from .conversions import (
    build_grpc_id,
    from_frame_to_grpc_frame,
    from_grpc_edge_tess_to_pd,
    from_grpc_edge_tess_to_raw_data,
    from_grpc_point_to_point3d,
    from_grpc_tess_to_pd,
    from_grpc_tess_to_raw_data,
    from_length_to_grpc_quantity,
    from_plane_to_grpc_plane,
    from_point3d_to_grpc_point,
    from_sketch_shapes_to_grpc_geometries,
    from_tess_options_to_grpc_tess_options,
    from_trimmed_curve_to_grpc_trimmed_curve,
    from_unit_vector_to_grpc_direction,
    serialize_tracked_command_response,
)


class GRPCBodyServiceV1(GRPCBodyService):
    """Body service for gRPC communication with the Geometry server.

    This class provides methods to create and manipulate bodies in the
    Geometry server using gRPC. It is specifically designed for the v1
    version of the Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):
        """Initialize the BodyService with the gRPC stub."""
        from ansys.api.discovery.v1.design.geometry.body_pb2_grpc import BodyStub
        from ansys.api.discovery.v1.design.geometry.face_pb2_grpc import FaceStub
        from ansys.api.discovery.v1.operations.edit_pb2_grpc import EditStub

        self.stub = BodyStub(channel)
        self.edit_stub = EditStub(channel)
        self.face_stub = FaceStub(channel)

    @protect_grpc
    def create_sphere_body(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            CreateSphereBodyRequest,
            CreateSphereBodyRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateSphereBodyRequest(
            request_data=[
                CreateSphereBodyRequestData(
                    name=kwargs["name"],
                    parent_id=build_grpc_id(kwargs["parent"]),
                    center=from_point3d_to_grpc_point(kwargs["center"]),
                    radius=from_length_to_grpc_quantity(kwargs["radius"]),
                )
            ]
        )

        # Call the gRPC service
        resp = self.stub.CreateSphereBody(request=request)

        # Return the response - formatted as a dictionary
        # Note: response.bodies is a repeated field, we return the first one
        body = resp.bodies[0]
        return {
            "id": body.id.id,
            "name": body.name,
            "master_id": body.master_id.id,
            "is_surface": body.is_surface,
        }

    @protect_grpc
    def create_extruded_body(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            CreateExtrudedBodyRequest,
            CreateExtrudedBodyRequestData,
        )

        # Apply direction (can be 1 or -1) to distance
        # distance_value = kwargs["distance"].value * kwargs["direction"]
        # distance=from_measurement_to_server_length(kwargs["distance"]) * kwargs["direction"],
        distance = Distance(kwargs["distance"].value * kwargs["direction"], kwargs["distance"].unit)

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateExtrudedBodyRequest(
            request_data=[
                CreateExtrudedBodyRequestData(
                    name=kwargs["name"],
                    parent_id=build_grpc_id(kwargs["parent_id"]),
                    plane=from_plane_to_grpc_plane(kwargs["sketch"].plane),
                    geometries=from_sketch_shapes_to_grpc_geometries(
                        kwargs["sketch"].plane, kwargs["sketch"].edges, kwargs["sketch"].faces
                    ),
                    distance=from_length_to_grpc_quantity(distance),
                )
            ]
        )

        # Call the gRPC service
        resp = self.stub.CreateExtrudedBody(request=request)

        # Return the response - formatted as a dictionary
        body = resp.bodies[0]
        return {
            "id": body.id.id,
            "name": body.name,
            "master_id": body.master_id.id,
            "is_surface": body.is_surface,
        }

    @protect_grpc
    def create_sweeping_profile_body(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            CreateSweepingProfileRequest,
            CreateSweepingProfileRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateSweepingProfileRequest(
            request_data=[
                CreateSweepingProfileRequestData(
                    name=kwargs["name"],
                    parent_id=build_grpc_id(kwargs["parent_id"]),
                    plane=from_plane_to_grpc_plane(kwargs["sketch"].plane),
                    geometries=from_sketch_shapes_to_grpc_geometries(
                        kwargs["sketch"].plane, kwargs["sketch"].edges, kwargs["sketch"].faces
                    ),
                    path=[from_trimmed_curve_to_grpc_trimmed_curve(tc) for tc in kwargs["path"]],
                )
            ]
        )

        # Call the gRPC service
        resp = self.stub.CreateSweepingProfile(request=request)

        # Return the response - formatted as a dictionary
        body = resp.bodies[0]
        return {
            "id": body.id.id,
            "name": body.name,
            "master_id": body.master_id.id,
            "is_surface": body.is_surface,
        }

    @protect_grpc
    def create_sweeping_chain(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            CreateSweepingChainRequest,
            CreateSweepingChainRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateSweepingChainRequest(
            request_data=[
                CreateSweepingChainRequestData(
                    name=kwargs["name"],
                    parent_id=build_grpc_id(kwargs["parent_id"]),
                    chain=[from_trimmed_curve_to_grpc_trimmed_curve(tc) for tc in kwargs["chain"]],
                    path=[from_trimmed_curve_to_grpc_trimmed_curve(tc) for tc in kwargs["path"]],
                )
            ]
        )

        # Call the gRPC service
        resp = self.stub.CreateSweepingChain(request=request)

        # Return the response - formatted as a dictionary
        body = resp.bodies[0]
        return {
            "id": body.id.id,
            "name": body.name,
            "master_id": body.master_id.id,
            "is_surface": body.is_surface,
        }

    @protect_grpc
    def sweep_with_guide(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            SweepWithGuideRequest,
            SweepWithGuideRequestData,
        )

        # Create request object - assumes all inputs are valid and of the proper type
        request = SweepWithGuideRequest(
            request_data=[
                SweepWithGuideRequestData(
                    name=sweep_item.name,
                    parent_id=build_grpc_id(sweep_item.parent_id),
                    plane=from_plane_to_grpc_plane(sweep_item.sketch.plane),
                    geometries=from_sketch_shapes_to_grpc_geometries(
                        sweep_item.sketch.plane,
                        sweep_item.sketch.edges,
                        sweep_item.sketch.faces,
                    ),
                    path=from_trimmed_curve_to_grpc_trimmed_curve(sweep_item.path),
                    guide=from_trimmed_curve_to_grpc_trimmed_curve(sweep_item.guide),
                    tight_tolerance=sweep_item.tight_tolerance,
                )
                for sweep_item in kwargs["sweep_data"]
            ],
        )

        # Call the gRPC service
        resp = self.edit_stub.SweepWithGuide(request=request)

        # Return the response - formatted as a dictionary
        return {
            "bodies": [
                {
                    "id": body.id.id,
                    "name": body.name,
                    "master_id": body.master_id.id,
                    "is_surface": body.is_surface,
                }
                for body in resp.bodies
            ]
        }

    @protect_grpc
    def create_extruded_body_from_face_profile(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            CreateExtrudedBodyFromFaceProfileRequest,
            CreateExtrudedBodyFromFaceProfileRequestData,
        )

        # Apply direction (can be 1 or -1) to distance
        # distance_value = kwargs["distance"].value * kwargs["direction"]
        distance = Distance(kwargs["distance"].value * kwargs["direction"], kwargs["distance"].unit)
        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateExtrudedBodyFromFaceProfileRequest(
            request_data=[
                CreateExtrudedBodyFromFaceProfileRequestData(
                    name=kwargs["name"],
                    parent_id=build_grpc_id(kwargs["parent_id"]),
                    face_id=build_grpc_id(kwargs["face_id"]),
                    distance=from_length_to_grpc_quantity(distance),
                )
            ]
        )

        # Call the gRPC service
        resp = self.stub.CreateExtrudedBodyFromFaceProfile(request=request)

        # Return the response - formatted as a dictionary
        body = resp.bodies[0]
        return {
            "id": body.id.id,
            "name": body.name,
            "master_id": body.master_id.id,
            "is_surface": body.is_surface,
        }

    @protect_grpc
    def create_extruded_body_from_loft_profiles(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.designmessages_pb2 import TrimmedCurveList
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            CreateExtrudedBodyFromLoftProfilesRequest,
            CreateExtrudedBodyFromLoftProfilesRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateExtrudedBodyFromLoftProfilesRequest(
            request_data=[
                CreateExtrudedBodyFromLoftProfilesRequestData(
                    name=kwargs["name"],
                    parent_id=build_grpc_id(kwargs["parent_id"]),
                    profiles=[
                        TrimmedCurveList(
                            curves=[from_trimmed_curve_to_grpc_trimmed_curve(tc) for tc in profile]
                        )
                        for profile in kwargs["profiles"]
                    ],
                    periodic=kwargs["periodic"],
                    ruled=kwargs["ruled"],
                )
            ]
        )

        # Call the gRPC service
        resp = self.stub.CreateExtrudedBodyFromLoftProfiles(request=request)

        # Return the response - formatted as a dictionary
        body = resp.bodies[0]
        return {
            "id": body.id.id,
            "name": body.name,
            "master_id": body.master_id.id,
            "is_surface": body.is_surface,
        }

    @protect_grpc
    def create_planar_body(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            CreatePlanarBodyRequest,
            CreatePlanarBodyRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreatePlanarBodyRequest(
            request_data=[
                CreatePlanarBodyRequestData(
                    name=kwargs["name"],
                    parent_id=build_grpc_id(kwargs["parent_id"]),
                    plane=from_plane_to_grpc_plane(kwargs["sketch"].plane),
                    geometries=from_sketch_shapes_to_grpc_geometries(
                        kwargs["sketch"].plane, kwargs["sketch"].edges, kwargs["sketch"].faces
                    ),
                )
            ]
        )

        # Call the gRPC service
        resp = self.stub.CreatePlanarBody(request=request)

        # Return the response - formatted as a dictionary
        body = resp.bodies[0]
        return {
            "id": body.id.id,
            "name": body.name,
            "master_id": body.master_id.id,
            "is_surface": body.is_surface,
        }

    @protect_grpc
    def create_body_from_face(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            CreateBodyFromFaceRequest,
            CreateBodyFromFaceRequestData,
        )

        # Create the request with inline request_data
        request = CreateBodyFromFaceRequest(
            request_data=[
                CreateBodyFromFaceRequestData(
                    name=kwargs["name"],
                    parent_id=build_grpc_id(kwargs["parent_id"]),
                    face_id=build_grpc_id(kwargs["face_id"]),
                )
            ]
        )

        # Call the gRPC service
        resp = self.stub.CreateFromFace(request=request)

        # Return the response - formatted as a dictionary
        body = resp.bodies[0]
        return {
            "id": body.id.id,
            "name": body.name,
            "master_id": body.master_id.id,
            "is_surface": body.is_surface,
        }

    @protect_grpc
    def create_surface_body(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            CreateSurfaceBodyRequest,
            CreateSurfaceBodyRequestData,
        )

        from ansys.geometry.core._grpc._services.v1.conversions import (
            from_trimmed_surface_to_grpc_trimmed_surface,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateSurfaceBodyRequest(
            request_data=[
                CreateSurfaceBodyRequestData(
                    name=kwargs["name"],
                    parent_id=build_grpc_id(kwargs["parent_id"]),
                    trimmed_surface=from_trimmed_surface_to_grpc_trimmed_surface(
                        kwargs["trimmed_surface"]
                    ),
                )
            ]
        )

        # Call the gRPC service
        resp = self.stub.CreateSurfaceBody(request=request)

        # Return the response - formatted as a dictionary
        body = resp.bodies[0]
        return {
            "id": body.id.id,
            "name": body.name,
            "master_id": body.master_id.id,
            "is_surface": body.is_surface,
        }

    @protect_grpc
    def create_surface_body_from_trimmed_curves(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            CreateSurfaceBodyFromTrimmedCurvesRequest,
            CreateSurfaceBodyFromTrimmedCurvesRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateSurfaceBodyFromTrimmedCurvesRequest(
            request_data=[
                CreateSurfaceBodyFromTrimmedCurvesRequestData(
                    name=kwargs["name"],
                    parent_id=build_grpc_id(kwargs["parent_id"]),
                    trimmed_curves=[
                        from_trimmed_curve_to_grpc_trimmed_curve(tc)
                        for tc in kwargs["trimmed_curves"]
                    ],
                )
            ]
        )

        # Call the gRPC service
        resp = self.stub.CreateSurfaceBodyFromTrimmedCurves(request=request)

        # Return the response - formatted as a dictionary
        body = resp.bodies[0]
        return {
            "id": body.id.id,
            "name": body.name,
            "master_id": body.master_id.id,
            "is_surface": body.is_surface,
        }

    @protect_grpc
    def translate(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            TranslateRequest,
            TranslateRequestData,
        )

        # Create the request with selection_ids, direction, and distance
        request = TranslateRequest(
            request_data=[
                TranslateRequestData(
                    id=build_grpc_id(id),
                    translation=from_unit_vector_to_grpc_direction(kwargs["direction"]),
                    distance=from_measurement_to_server_length(kwargs["distance"]),
                )
                for id in kwargs["ids"]
            ]
        )

        # Call the gRPC service
        self.edit_stub.Translate(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def delete(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commonmessages_pb2 import MultipleEntitiesRequest

        # Create the request with MultipleEntitiesRequest
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])

        # Call the gRPC service
        self.stub.Delete(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def is_suppressed(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commonmessages_pb2 import MultipleEntitiesRequest

        # Create the request with MultipleEntitiesRequest
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])

        # Call the gRPC service
        resp = self.stub.GetIsSuppressed(request=request)

        # Return the response - formatted as a dictionary (get first value from map)
        result = next(iter(resp.result.values())) if resp.result else False
        return {"result": result}

    @protect_grpc
    def get_color(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commonmessages_pb2 import (
            MultipleEntitiesRequest,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])

        # Call the gRPC service
        resp = self.stub.GetColor(request=request)

        # Return the response - formatted as a dictionary (get first color from map)
        color = next(iter(resp.colors.values())) if resp.colors else None
        return {"color": color}

    @protect_grpc
    def set_color(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.designmessages_pb2 import (
            SetColorRequest,
            SetColorRequestData,
        )

        # Create the request with repeated request_data
        request = SetColorRequest(
            request_data=[
                SetColorRequestData(
                    id=build_grpc_id(kwargs["id"]),
                    color=kwargs["color"],
                )
            ]
        )

        # Call the gRPC service
        self.stub.SetColor(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def get_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commonmessages_pb2 import MultipleEntitiesRequest

        # Create the request with MultipleEntitiesRequest
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])

        # Call the gRPC service
        resp = self.stub.GetFaces(request=request)

        # Get the first response data from the array
        response_data = resp.response_data[0]

        # Return the response - formatted as a dictionary
        return {
            "faces": [
                {
                    "id": face.id.id,
                    "surface_type": face.surface_type,
                    "is_reversed": face.is_reversed,
                }
                for face in response_data.faces
            ]
        }

    @protect_grpc
    def get_edges(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commonmessages_pb2 import MultipleEntitiesRequest

        # Create the request with MultipleEntitiesRequest
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])

        # Call the gRPC service
        resp = self.stub.GetEdges(request=request)

        # Get the first response data from the array
        response_data = resp.response_data[0]

        # Return the response - formatted as a dictionary
        return {
            "edges": [
                {
                    "id": edge.id.id,
                    "curve_type": edge.curve_type,
                    "is_reversed": edge.is_reversed,
                }
                for edge in response_data.edges
            ]
        }

    @protect_grpc
    def get_vertices(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commonmessages_pb2 import MultipleEntitiesRequest

        # Create the request with MultipleEntitiesRequest
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])

        # Call the gRPC service
        resp = self.stub.GetVertices(request=request)

        # Get the first response data from the array
        response_data = resp.response_data[0]

        # Return the response - formatted as a dictionary
        return {
            "vertices": [
                {
                    "id": vertex.id.id,
                    "position": from_grpc_point_to_point3d(vertex.position),
                }
                for vertex in response_data.vertices
            ]
        }

    @protect_grpc
    def get_volume(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commonmessages_pb2 import MultipleEntitiesRequest

        from .conversions import from_grpc_volume_to_volume

        # Create the request with MultipleEntitiesRequest
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])

        # Call the gRPC service
        resp = self.stub.GetVolume(request=request)

        # Get the first response data from the array
        response_data = resp.response_data[0]

        # Return the response - formatted as a dictionary
        return {"volume": from_grpc_volume_to_volume(response_data.volume)}

    @protect_grpc
    def get_bounding_box(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commonmessages_pb2 import MultipleEntitiesRequest
        from ansys.api.discovery.v1.design.designmessages_pb2 import (
            GetBoundingBoxRequest,
            GetBoundingBoxRequestData,
        )

        # Create the request to the proper method depending on tight tolerance
        if kwargs.get("tight"):
            request = GetBoundingBoxRequest(
                request_data=[
                    GetBoundingBoxRequestData(
                        id=build_grpc_id(kwargs["id"]),
                        tight_tolerance=kwargs.get("tight", False),
                    )
                ]
            )

            resp = self.stub.GetTightBoundingBox(request).response_data[0]
        else:
            request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])
            resp = self.stub.GetBoundingBox(request).response_data[0]

        # Return the response - formatted as a dictionary
        return {
            "min": from_grpc_point_to_point3d(resp.box.min),
            "max": from_grpc_point_to_point3d(resp.box.max),
            "center": from_grpc_point_to_point3d(resp.box.center),
        }

    @protect_grpc
    def set_assigned_material(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            SetAssignedCADMaterialRequest,
            SetAssignedCADMaterialRequestData,
        )

        from ansys.geometry.core._grpc._services.v1.conversions import (
            from_material_to_grpc_material,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = SetAssignedCADMaterialRequest(
            request_data=[
                SetAssignedCADMaterialRequestData(
                    id=build_grpc_id(kwargs["id"]),
                    material=from_material_to_grpc_material(kwargs["material"]),
                )
            ]
        )

        # Call the gRPC service
        self.stub.SetAssignedCADMaterial(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def get_assigned_material(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commonmessages_pb2 import MultipleEntitiesRequest
        from ansys.api.discovery.v1.design.designmessages_pb2 import MaterialEntity

        from ansys.geometry.core._grpc._services.v1.conversions import (
            from_grpc_material_to_material,
        )

        # Create the request with MultipleEntitiesRequest
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])

        # Call the gRPC service
        resp = self.stub.GetAssignedCADMaterial(request=request)

        # Return the response - formatted as a dictionary
        # Get the first response data from the array
        response_data = resp.response_data[0] if resp.response_data else None
        grpc_material = (
            response_data.material
            if response_data and response_data.HasField("material")
            else MaterialEntity()
        )
        return {"material": from_grpc_material_to_material(grpc_material)}

    @protect_grpc
    def remove_assigned_material(self, **kwargs):  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            RemoveAssignedCADMaterialRequest,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = RemoveAssignedCADMaterialRequest(ids=[build_grpc_id(id) for id in kwargs["ids"]])

        # Call the gRPC service
        resp = self.stub.RemoveAssignedCADMaterial(request=request)

        # Return the response - formatted as a dictionary
        return {"successfully_removed": [id.id for id in resp.successfully_removed_ids]}

    @protect_grpc
    def set_name(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.designmessages_pb2 import (
            SetDesignEntityNameRequest,
            SetDesignEntityNameRequestData,
        )

        # Create the request with repeated request_data
        request = SetDesignEntityNameRequest(
            request_data=[
                SetDesignEntityNameRequestData(
                    id=build_grpc_id(kwargs["id"]),
                    name=kwargs["name"],
                )
            ]
        )

        # Call the gRPC service
        self.stub.SetName(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def set_fill_style(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            SetFillStyleRequest,
            SetFillStyleRequestData,
        )

        # Create the request with request_data
        request = SetFillStyleRequest(
            request_data=[
                SetFillStyleRequestData(
                    id=build_grpc_id(kwargs["id"]), fill_style=kwargs["fill_style"].value
                )
            ]
        )

        # Call the gRPC service
        self.stub.SetFillStyle(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def set_suppressed(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import SetSuppressedRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = SetSuppressedRequest(
            body_ids=[build_grpc_id(body_id) for body_id in kwargs["bodies"]],
            is_suppressed=kwargs["is_suppressed"],
        )

        # Call the gRPC service
        self.stub.SetSuppressed(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def rotate(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            RotateOptions,
            RotateRequest,
            RotateRequestData,
        )

        from .conversions import from_angle_to_grpc_quantity

        # Create options
        options = RotateOptions(
            angle=from_angle_to_grpc_quantity(kwargs["angle"]),
            axis_origin=from_point3d_to_grpc_point(kwargs["axis_origin"]),
            axis_direction=from_unit_vector_to_grpc_direction(kwargs["axis_direction"]),
        )

        # Create the request with selection_ids, axis, and angle
        request = RotateRequest(
            request_data=[RotateRequestData(id=build_grpc_id(kwargs["id"]), options=options)]
        )

        # Call the gRPC service
        self.edit_stub.Rotate(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def scale(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import ScaleRequest, ScaleRequestData

        # Create the request with inline request_data
        request = ScaleRequest(
            request_data=[
                ScaleRequestData(
                    object_ids=[build_grpc_id(kwargs["id"])],
                    scale=kwargs["value"],
                )
            ]
        )

        # Call the gRPC service
        self.edit_stub.Scale(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def mirror(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import MirrorRequest, MirrorRequestData

        # Create the request - assumes all inputs are valid and of the proper type
        request = MirrorRequest(
            request_data=[
                MirrorRequestData(
                    ids=[build_grpc_id(kwargs["id"])],
                    mirror_plane=from_plane_to_grpc_plane(kwargs["plane"]),
                )
            ]
        )

        # Call the gRPC service
        self.edit_stub.Mirror(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def map(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import MapBodyRequest, MapBodyRequestData

        # Create the request - assumes all inputs are valid and of the proper type
        request = MapBodyRequest(
            request_data=[
                MapBodyRequestData(
                    id=build_grpc_id(kwargs["id"]),
                    frame=from_frame_to_grpc_frame(kwargs["frame"]),
                )
            ]
        )

        # Call the gRPC service
        self.edit_stub.MapBody(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def get_collision(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            GetCollisionRequest,
            GetCollisionRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = GetCollisionRequest(
            request_data=[
                GetCollisionRequestData(
                    body_1_id=build_grpc_id(kwargs["id"]),
                    body_2_id=build_grpc_id(kwargs["other_id"]),
                )
            ]
        )

        # Call the gRPC service
        resp = self.stub.GetCollision(request=request)

        # Return the response - formatted as a dictionary
        return {"collision_type": resp.response_data[0].collision}

    @protect_grpc
    def copy(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            CreateDuplicateBodyRequest,
            DuplicateBodyRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateDuplicateBodyRequest(
            request_data=[
                DuplicateBodyRequestData(
                    id=build_grpc_id(kwargs["id"]),
                    parent=build_grpc_id(kwargs["parent_id"]),
                    name=kwargs["name"],
                )
            ]
        )

        # Call the gRPC service
        resp = self.stub.CreateDuplicate(request=request)

        # Return the response - formatted as a dictionary
        response_data = resp.response_data[0]
        body = response_data.body
        return {"id": body.id.id, "master_id": body.master_id.id, "name": body.name}

    @protect_grpc
    def get_tesellation(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            GetTessellationRequest,
            GetTessellationRequestData,
        )

        tess_map = {}
        resp = []  # For compatibility with stream response
        try:
            request = GetTessellationRequest(
                request_data=[GetTessellationRequestData(id=build_grpc_id(kwargs["id"]))]
            )
            resp = [self.stub.GetTessellation(request=request)]
        except grpc.RpcError:
            request = GetTessellationRequest(
                request_data=[GetTessellationRequestData(id=build_grpc_id(kwargs["id"]))]
            )
            resp = self.stub.GetTessellation(request=request)

        for elem in resp:
            for face_id, face_tess in elem.face_tessellation.items():
                tess_map[face_id] = (
                    from_grpc_tess_to_raw_data(face_tess)
                    if kwargs["raw_data"]
                    else from_grpc_tess_to_pd(face_tess)
                )

        return {"tessellation": tess_map}

    @protect_grpc
    def get_tesellation_with_options(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            GetTessellationRequest,
            GetTessellationRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = GetTessellationRequest(
            request_data=[
                GetTessellationRequestData(
                    id=build_grpc_id(kwargs["id"]),
                    options=from_tess_options_to_grpc_tess_options(kwargs["options"]),
                    include_faces=kwargs["include_faces"],
                    include_edges=kwargs["include_edges"],
                )
            ]
        )

        tess_map = {}
        resp = []  # For compatibility with stream response
        try:
            resp = [self.stub.GetTessellation(request=request)]
        except grpc.RpcError:
            resp = self.stub.GetTessellationStream(request=request)

        for elem in resp:
            for face_id, face_tess in elem.face_tessellation.items():
                tess_map[face_id] = (
                    from_grpc_tess_to_raw_data(face_tess)
                    if kwargs["raw_data"]
                    else from_grpc_tess_to_pd(face_tess)
                )
        return {"tessellation": tess_map}

    @protect_grpc
    def boolean(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import BooleanRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = BooleanRequest(
            body1_id=build_grpc_id(kwargs["target"]),
            tool_body_ids=[build_grpc_id(id) for id in kwargs["other"]],
            method=kwargs["method"],
            keep_other=kwargs["keep_other"],
        )

        # Call the gRPC service
        response = self.edit_stub.Boolean(request=request)

        if not response.tracked_command_response.command_response.success:
            raise ValueError(f"Boolean operation failed: {kwargs['err_msg']}")

        serialized_response = serialize_tracked_command_response(response.tracked_command_response)

        # Return the response - formatted as a dictionary
        return {"tracker_response": serialized_response}

    @protect_grpc
    def combine(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            CombineIntersectBodiesRequest,
            CombineIntersectBodiesRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = CombineIntersectBodiesRequest(
            request_data=[
                CombineIntersectBodiesRequestData(
                    target_selection_ids=[build_grpc_id(kwargs["target"])],
                    tool_selection_ids=[build_grpc_id(id) for id in kwargs["other"]],
                    keep_cutter=kwargs["keep_other"],
                    subtract_from_target=True,
                    transfer_named_selections=kwargs["transfer_named_selections"],
                )
            ]
        )

        # Call the gRPC service
        response = self.edit_stub.CombineIntersectBodies(request=request)

        serialized_response = serialize_tracked_command_response(
            response.tracked_command_response
        )  # Local alias  # noqa: E501
        if not response.tracked_command_response.command_response.success:
            raise ValueError(f"Boolean operation failed: {kwargs['err_msg']}")

        # Return the response - formatted as a dictionary
        return {"tracker_response": serialized_response}

    @protect_grpc
    def split_body(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            SplitBodyRequest,
            SplitBodyRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = SplitBodyRequest(
            request_data=[
                SplitBodyRequestData(
                    selection_ids=[build_grpc_id(id) for id in kwargs["body_ids"]],
                    split_by_plane=(
                        from_plane_to_grpc_plane(kwargs["plane"]) if kwargs["plane"] else None
                    ),
                    split_by_slicer_ids=[build_grpc_id(id) for id in kwargs["slicer_ids"]],
                    split_by_face_ids=[build_grpc_id(id) for id in kwargs["face_ids"]],
                    extend_surfaces=kwargs["extend_surfaces"],
                )
            ]
        )

        # Call the gRPC service
        resp = self.edit_stub.SplitBodies(request=request)
        tracked_response = serialize_tracked_command_response(resp.tracked_command_response)

        # Return the response - formatted as a dictionary
        return {
            "success": resp.tracked_command_response.command_response.success,
            "tracked_response": tracked_response,
        }

    @protect_grpc
    def create_body_from_loft_profiles_with_guides(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.designmessages_pb2 import TrimmedCurveList
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            CreateBodyFromLoftWithGuidesRequest,
            CreateBodyFromLoftWithGuidesRequestData,
        )

        # Create request object - assumes all inputs are valid and of the proper type
        request = CreateBodyFromLoftWithGuidesRequest(
            request_data=[
                CreateBodyFromLoftWithGuidesRequestData(
                    name=kwargs["name"],
                    parent_id=build_grpc_id(kwargs["parent_id"]),
                    profiles=[
                        TrimmedCurveList(
                            curves=[from_trimmed_curve_to_grpc_trimmed_curve(tc) for tc in profile]
                        )
                        for profile in kwargs["profiles"]
                    ],
                    guides=TrimmedCurveList(
                        curves=[
                            from_trimmed_curve_to_grpc_trimmed_curve(tc) for tc in kwargs["guides"]
                        ]
                    ),
                )
            ]
        )

        # Call the gRPC service
        response = self.stub.CreateFromLoftWithGuides(request)

        # Return the response - formatted as a dictionary
        new_body = response.created_bodies[0]
        return {
            "id": new_body.id.id,
            "name": new_body.name,
            "master_id": new_body.master_id.id,
            "is_surface": new_body.is_surface,
        }

    @protect_grpc
    def combine_merge(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            CombineMergeBodiesRequest,
            CombineMergeBodiesRequestData,
        )

        # Create the request with inline request_data
        request = CombineMergeBodiesRequest(
            request_data=[
                CombineMergeBodiesRequestData(
                    target_selection_ids=[build_grpc_id(body_id) for body_id in kwargs["body_ids"]]
                )
            ]
        )

        # Call the gRPC service
        _ = self.edit_stub.CombineMergeBodies(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def assign_midsurface_thickness(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            SetMidSurfaceThicknessRequest,
            SetMidSurfaceThicknessRequestData,
        )

        request = SetMidSurfaceThicknessRequest(
            request_data=[
                SetMidSurfaceThicknessRequestData(
                    ids=[build_grpc_id(id) for id in kwargs["ids"]],
                    thickness=from_length_to_grpc_quantity(kwargs["thickness"]),
                )
            ]
        )

        _ = self.stub.SetMidSurfaceThickness(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def assign_midsurface_offset(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            SetMidSurfaceOffsetTypeRequest,
            SetMidSurfaceOffsetTypeRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = SetMidSurfaceOffsetTypeRequest(
            request_data=[
                SetMidSurfaceOffsetTypeRequestData(
                    ids=[build_grpc_id(id) for id in kwargs["ids"]],
                    offset_type=kwargs["offset_type"].value,
                )
            ]
        )

        # Call the gRPC service
        _ = self.stub.SetMidSurfaceOffsetType(request=request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def shell(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import ShellRequest, ShellRequestData

        # Create the request with request_data
        request = ShellRequest(
            request_data=[
                ShellRequestData(
                    selection_id=build_grpc_id(kwargs["id"]),
                    offset=from_length_to_grpc_quantity((kwargs["offset"])),
                )
            ]
        )

        # Call the gRPC service
        response = self.edit_stub.Shell(request=request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.tracked_command_response.command_response.success,
            "tracked_command_response": response.tracked_command_response,
        }

    @protect_grpc
    def remove_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.face_pb2 import (
            RemoveFacesRequest,
            RemoveFacesRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        offset_quantity = from_length_to_grpc_quantity(kwargs["offset"])
        request_data = RemoveFacesRequestData(
            selection_ids=[build_grpc_id(id) for id in kwargs["face_ids"]],
            offset=offset_quantity,
        )
        request = RemoveFacesRequest(
            request_data=[request_data],
            offset=offset_quantity,
        )

        # Call the gRPC service
        response = self.face_stub.Remove(request=request)

        # Return the response - formatted as a dictionary
        return {
            "success": response.tracked_command_response.command_response.success,
            "tracked_command_response": response.tracked_command_response,
        }

    @protect_grpc
    def imprint_curves(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            ImprintCurvesRequest,
            ImprintCurvesRequestData,
        )

        # Convert sketch and trimmed curves to gRPC format
        sketch = kwargs["sketch"]
        curves = (
            from_sketch_shapes_to_grpc_geometries(sketch.plane, sketch.edges, sketch.faces)
            if sketch
            else None
        )

        # Convert trimmed curves to gRPC format
        trimmed_curves = []
        if kwargs.get("tc"):
            trimmed_curves = [from_trimmed_curve_to_grpc_trimmed_curve(tc) for tc in kwargs["tc"]]

        # Create the request using request_data pattern
        request = ImprintCurvesRequest(
            request_data=[
                ImprintCurvesRequestData(
                    body_id=build_grpc_id(kwargs["id"]),
                    curves=curves,
                    face_ids=[build_grpc_id(id) for id in kwargs["face_ids"]],
                    plane=from_plane_to_grpc_plane(sketch.plane) if sketch else None,
                    trimmed_curves=trimmed_curves,
                )
            ]
        )

        # Call the gRPC service
        response = self.edit_stub.ImprintCurves(request=request).response_data[0]

        # Return the response - formatted as a dictionary
        return {
            "edges": [
                {
                    "id": edge.id.id,
                    "curve_type": edge.curve_type,
                    "is_reversed": edge.is_reversed,
                }
                for edge in response.created_entities.edges
            ],
            "faces": [
                {
                    "id": face.id.id,
                    "surface_type": face.surface_type,
                    "is_reversed": face.is_reversed,
                }
                for face in response.created_entities.faces
            ],
        }

    @protect_grpc
    def project_curves(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            ProjectCurvesRequest,
            ProjectCurvesRequestData,
        )

        # Convert sketch to geometries
        sketch = kwargs["sketch"]
        curves = from_sketch_shapes_to_grpc_geometries(sketch.plane, sketch.edges, sketch.faces)

        # Create the request using ProjectCurvesRequestData
        request = ProjectCurvesRequest(
            request_data=[
                ProjectCurvesRequestData(
                    body_id=build_grpc_id(kwargs["id"]),
                    curves=curves,
                    direction=from_unit_vector_to_grpc_direction(kwargs["direction"]),
                    closest_face=kwargs["closest_face"],
                    plane=from_plane_to_grpc_plane(sketch.plane),
                )
            ]
        )

        # Call the gRPC service
        resp = self.edit_stub.ProjectCurves(request=request)

        # Return the response - formatted as a dictionary
        return {
            "faces": [
                {
                    "id": face.id.id,
                    "surface_type": face.surface_type,
                    "is_reversed": face.is_reversed,
                }
                for face in resp.response_data[0].faces
            ]
        }

    @protect_grpc
    def imprint_projected_curves(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.operations.edit_pb2 import (
            ImprintProjectedCurvesRequest,
            ImprintProjectedCurvesRequestData,
        )

        # Convert sketch to geometries
        sketch = kwargs["sketch"]
        curves = from_sketch_shapes_to_grpc_geometries(sketch.plane, sketch.edges, sketch.faces)

        # Create the request using ImprintProjectedCurvesRequestData
        request = ImprintProjectedCurvesRequest(
            request_data=[
                ImprintProjectedCurvesRequestData(
                    body_id=build_grpc_id(kwargs["id"]),
                    curves=curves,
                    direction=from_unit_vector_to_grpc_direction(kwargs["direction"]),
                    closest_face=kwargs["closest_face"],
                    plane=from_plane_to_grpc_plane(sketch.plane),
                )
            ]
        )

        # Call the gRPC service
        resp = self.edit_stub.ImprintProjectedCurves(request=request).response_data[0]

        # Return the response - formatted as a dictionary
        return {
            "faces": [
                {
                    "id": face.id.id,
                    "surface_type": face.surface_type,
                    "is_reversed": face.is_reversed,
                }
                for face in resp.created_entities.faces
            ],
        }

    @protect_grpc
    def get_full_tessellation(self, **kwargs):  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            GetTessellationRequest,
            GetTessellationRequestData,
        )

        # Create options
        options = kwargs["options"] if kwargs["options"] else None

        # Create the request - assumes all inputs are valid and of the proper type
        request = GetTessellationRequest(
            request_data=[
                GetTessellationRequestData(
                    id=build_grpc_id(kwargs["id"]),
                    options=from_tess_options_to_grpc_tess_options(options) if options else None,
                    include_faces=kwargs["include_faces"],
                    include_edges=kwargs["include_edges"],
                )
            ]
        )

        # Call the gRPC service - using streaming
        resp = self.stub.GetTessellationStream(request=request)

        # Return the response - formatted as a dictionary
        tess_map = {}
        for elem in resp:
            for face_id, face_tess in elem.response_data[0].face_tessellation.items():
                tess_map[face_id] = (
                    from_grpc_tess_to_raw_data(face_tess)
                    if kwargs["raw_data"]
                    else from_grpc_tess_to_pd(face_tess)
                )
            for edge_id, edge_tess in elem.response_data[0].edge_tessellation.items():
                tess_map[edge_id] = (
                    from_grpc_edge_tess_to_raw_data(edge_tess)
                    if kwargs["raw_data"]
                    else from_grpc_edge_tess_to_pd(edge_tess)
                )

        return {"tessellation": tess_map}

    @protect_grpc
    def copy_faces(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.body_pb2 import (
            CopyFacesRequest,
            CopyFacesRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = CopyFacesRequest(
            request_data=[
                CopyFacesRequestData(
                    face_ids=[build_grpc_id(id) for id in kwargs["face_ids"]],
                    parent_id=build_grpc_id(kwargs["parent_id"]),
                    name=kwargs["name"],
                )
            ]
        )

        # Call the gRPC service
        resp = self.stub.CopyFaces(request=request)

        # Return the response - formatted as a dictionary
        body = resp.bodies[0]
        return {
            "id": body.id.id,
            "name": body.name,
            "master_id": body.master_id.id,
            "is_surface": body.is_surface,
        }



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module for creating and managing datum planes."""

from typing import TYPE_CHECKING

from ansys.geometry.core.math.plane import Plane
from ansys.geometry.core.math.point import Point3D
from ansys.geometry.core.misc.checks import min_backend_version

if TYPE_CHECKING:  # pragma: no cover
    from ansys.geometry.core.designer.component import Component


class DatumPlane:
    """Provides for creating datum planes in components.

    Parameters
    ----------
    id : str
        Server-defined ID for the datum plane.
    name : str
        User-defined label for the datum plane.
    plane : Plane
        Plane constituting the datum plane.
    parent_component : Component
        Parent component to place the new datum plane under within the design assembly.
    """

    def __init__(self, id: str, name: str, plane: Plane, parent_component: "Component"):
        """Initialize the ``DatumPlane`` class."""
        self._id = id
        self._name = name
        self._value = plane
        self._parent_component = parent_component
        self._is_alive = True

    @property
    def id(self) -> str:
        """ID of the datum plane."""
        return self._id

    @property
    def name(self) -> str:
        """Name of the datum plane."""
        return self._name

    @property
    @min_backend_version(27, 1, 0)
    def value(self) -> Plane:
        """Plane constituting the datum plane."""
        return self._value

    @property
    def parent_component(self) -> "Component":
        """Parent component of the datum plane."""
        return self._parent_component

    @property
    def _grpc_client(self):
        """Access to the gRPC client through the parent component."""
        return self._parent_component._grpc_client

    @property
    def is_alive(self) -> bool:
        """Check if the datum plane is still present on the server."""
        return self._is_alive

    def evaluate(self, u: float, v: float) -> Point3D:
        """Evaluate the plane at UV parametric coordinates.

        This method converts 2D parametric coordinates (u, v) to a 3D Cartesian point
        on the plane using the parametric equation:
        P(u, v) = origin + u * direction_x + v * direction_y

        Parameters
        ----------
        u : float
            First parametric coordinate along the plane's x-direction.
        v : float
            Second parametric coordinate along the plane's y-direction.

        Returns
        -------
        Point3D
            3D Cartesian point on the plane at the given UV coordinates.
        """
        return self.value.origin + u * self.value.direction_x + v * self.value.direction_y

    def __repr__(self):
        """Represent the datum plane as a string."""
        lines = [f"ansys.geometry.core.design.DatumPlane {hex(id(self))}"]
        lines.append(f"  Name                 : {self.name}")
        lines.append(f"  Datum Plane          : {self.value}")
        return "\n".join(lines)



# Copyright (C) 2023 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Module containing the components service implementation for v1."""

import grpc

from ansys.geometry.core.errors import protect_grpc

from ..base.components import GRPCComponentsService
from .conversions import (
    build_grpc_id,
    from_angle_to_grpc_quantity,
    from_grpc_matrix_to_matrix,
    from_point3d_to_grpc_point,
    from_unit_vector_to_grpc_direction,
)


class GRPCComponentsServiceV1(GRPCComponentsService):
    """Components service for gRPC communication with the Geometry server.

    This class provides methods to call components in the
    Geometry server using gRPC. It is specifically designed for the v1
    version of the Geometry API.

    Parameters
    ----------
    channel : grpc.Channel
        The gRPC channel to the server.
    """

    @protect_grpc
    def __init__(self, channel: grpc.Channel):  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.component_pb2_grpc import ComponentStub

        self.stub = ComponentStub(channel)

    @protect_grpc
    def create(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.component_pb2 import (
            CreateComponentData,
            CreateRequest,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = CreateRequest(
            components=[
                CreateComponentData(
                    name=kwargs["name"],
                    parent_id=build_grpc_id(kwargs["parent_id"]),
                    template_id=build_grpc_id(kwargs["template_id"]),
                    instance_name=kwargs["instance_name"],
                )
            ]
        )

        # Call the gRPC service
        response = self.stub.Create(request)

        # Return the response - formatted as a dictionary
        # Note: response.components is a repeated field, we return the first one
        component = response.components[0]
        return {
            "id": component.id.id,
            "name": component.name,
            "instance_name": component.instance_name,
            "template": kwargs["template_id"],  # template_id from input
            "component": component,
            "component_master_id": component.master_id.id,
            "component_part_master_id": component.part_master.id.id,
        }

    @protect_grpc
    def set_name(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.designmessages_pb2 import (
            SetDesignEntityNameRequest,
            SetDesignEntityNameRequestData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = SetDesignEntityNameRequest(
            request_data=[
                SetDesignEntityNameRequestData(
                    id=build_grpc_id(kwargs["id"]),
                    name=kwargs["name"],
                )
            ]
        )

        # Call the gRPC service
        _ = self.stub.SetName(request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def set_placement(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.component_pb2 import (
            PlacementData,
            SetPlacementRequest,
        )

        # Create the direction and point objects
        translation = (
            from_unit_vector_to_grpc_direction(kwargs["translation"])
            if kwargs["translation"] is not None
            else None
        )
        origin = (
            from_point3d_to_grpc_point(kwargs["rotation_axis_origin"])
            if kwargs["rotation_axis_origin"] is not None
            else None
        )
        direction = (
            from_unit_vector_to_grpc_direction(kwargs["rotation_axis_direction"])
            if kwargs["rotation_axis_direction"] is not None
            else None
        )

        # Create the request with repeated ids and placements
        request = SetPlacementRequest(
            ids=[build_grpc_id(kwargs["id"])],
            placements=[
                PlacementData(
                    translation=translation,
                    rotation_axis_origin=origin,
                    rotation_axis_direction=direction,
                    rotation_angle=from_angle_to_grpc_quantity(kwargs["rotation_angle"]),
                )
            ],
        )

        # Call the gRPC service
        response = self.stub.SetPlacement(request)

        # Return the response - formatted as a dictionary
        # Note: response.matrices is a map<string, Matrix>
        # Get the matrix for our component ID
        matrix_value = response.matrices.get(kwargs["id"])
        return {"matrix": from_grpc_matrix_to_matrix(matrix_value) if matrix_value else None}

    @protect_grpc
    def set_shared_topology(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.component_pb2 import (
            SetSharedTopologyRequest,
            SharedTopologyData,
        )

        # Create the request - assumes all inputs are valid and of the proper type
        request = SetSharedTopologyRequest(
            shared_topologies=[
                SharedTopologyData(
                    id=build_grpc_id(kwargs["id"]),
                    share_type=kwargs["share_type"].value,
                )
            ]
        )

        # Call the gRPC service
        _ = self.stub.SetSharedTopology(request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def delete(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.commonmessages_pb2 import MultipleEntitiesRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = MultipleEntitiesRequest(ids=[build_grpc_id(kwargs["id"])])

        # Call the gRPC service
        _ = self.stub.Delete(request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def import_groups(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.component_pb2 import ImportGroupsRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = ImportGroupsRequest(
            id=build_grpc_id(kwargs["id"]),
        )

        # Call the gRPC service
        _ = self.stub.ImportGroups(request)

        # Return the response - formatted as a dictionary
        return {}

    @protect_grpc
    def make_independent(self, **kwargs) -> dict:  # noqa: D102
        from ansys.api.discovery.v1.design.geometry.component_pb2 import MakeIndependentRequest

        # Create the request - assumes all inputs are valid and of the proper type
        request = MakeIndependentRequest(
            ids=[build_grpc_id(id) for id in kwargs["ids"]],
        )

        # Call the gRPC service
        _ = self.stub.MakeIndependent(request)

        # Return the response - formatted as a dictionary
        return {}
